# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query Provider mixin methods."""
import re
from collections import abc
from typing import Dict, Iterable, List, NamedTuple, Optional, Pattern, Protocol, Union

from ..._version import VERSION
from ...common.utility.package import delayed_import
from ..drivers.driver_base import DriverBase
from .query_defns import DataEnvironment
from .query_source import QuerySource
from .query_store import QueryStore

__version__ = VERSION
__author__ = "Ian Hellen"

query_browser = delayed_import("msticpy.vis.query_browser", "browse_queries")


# pylint: disable=too-few-public-methods
class QueryProviderProtocol(Protocol):
    """Protocol for required properties of QueryProvider class."""

    _query_provider: DriverBase
    query_store: QueryStore

    def _add_query_functions(self):
        pass


class QueryParam(NamedTuple):
    """
    Named tuple for custom query parameters.

    name and data_type are mandatory.
    description and default are optional.

    """

    name: str
    data_type: str
    description: Optional[str] = None
    default: Optional[str] = None


# pylint: disable=super-init-not-called
class QueryProviderUtilsMixin(QueryProviderProtocol):
    """Mixin utility methods for QueryProvider class."""

    create_param = QueryParam

    @property
    def connected(self) -> bool:
        """
        Return True if the provider is connected.

        Returns
        -------
        bool
            True if the provider is connected.

        """
        return self._query_provider.connected

    @property
    def connection_string(self) -> str:
        """
        Return provider connection string.

        Returns
        -------
        str
            Provider connection string.

        """
        return self._query_provider.current_connection

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._query_provider.schema

    @property
    def schema_tables(self) -> List[str]:
        """
        Return list of tables in the data schema of the connection.

        Returns
        -------
        List[str]
            Tables in the of current connection.

        """
        return list(self._query_provider.schema.keys())

    @property
    def instance(self) -> Optional[str]:
        """
        Return instance name, if any for provider.

        Returns
        -------
        Optional[str]
            The instance name or None for drivers that do not
            support multiple instances.

        """
        return self._query_provider.instance

    def import_query_file(self, query_file: str):
        """
        Import a yaml data source definition.

        Parameters
        ----------
        query_file : str
            Path to the file to import

        """
        self.query_store.import_file(query_file)
        self._add_query_functions()

    def driver_help(self):
        """Display help for the driver."""
        print(self._query_provider.__doc__)

    @classmethod
    def list_data_environments(cls) -> List[str]:
        """
        Return list of current data environments.

        Returns
        -------
        List[str]
            List of current data environments

        """
        # pylint: disable=not-an-iterable
        return [
            de
            for de in dir(DataEnvironment)
            if de != "Unknown" and not de.startswith("_")
        ]
        # pylint: enable=not-an-iterable

    def list_queries(self, substring: Optional[str] = None) -> List[str]:
        """
        Return list of family.query in the store.

        Parameters
        ----------
        substring : Optional[str]
            Optional pattern - will return only queries matching the pattern,
            default None.

        Returns
        -------
        List[str]
            List of queries

        """
        if substring:
            return list(
                filter(
                    lambda x: substring in x.lower(),  # type: ignore
                    self.query_store.query_names,
                )
            )
        return list(self.query_store.query_names)

    def search(
        self,
        search: Union[str, Iterable[str]] = None,
        table: Union[str, Iterable[str]] = None,
        param: Union[str, Iterable[str]] = None,
        ignore_case: bool = True,
    ) -> List[str]:
        """
        Search queries for match properties.

        Parameters
        ----------
        search : Union[str, Iterable[str]], optional
            String or iterable of search terms to match on
            any property of the query, by default None.
            The properties include: name, description, table,
            parameter names and query_text.
        table : Union[str, Iterable[str]], optional
            String or iterable of search terms to match on
            the query table name, by default None
        param : Union[str, Iterable[str]], optional
            String or iterable of search terms to match on
            the query parameter names, by default None
        ignore_case : bool
            Use case-insensitive search, default is True.

        Returns
        -------
        List[str]
            A list of matched queries

        Notes
        -----
        Search terms are treated as regular expressions.
        Supplying multiple parameters returns the intersection
        of matches for each category. For example:
        `qry_prov.search(search="account", table="syslog")` will
        match queries that have a table parameter of "syslog" AND
        have the term "Account" somewhere in the query properties.

        """
        if not (search or table or param):
            return []

        glob_searches = _normalize_to_regex(search, ignore_case)
        table_searches = _normalize_to_regex(table, ignore_case)
        param_searches = _normalize_to_regex(param, ignore_case)
        search_hits: List[str] = []
        for query, search_data in self.query_store.search_items.items():
            glob_match = (not glob_searches) or any(
                re.search(term, prop)
                for term in glob_searches
                for prop in search_data.values()
            )
            table_match = (not table_searches) or any(
                re.search(term, search_data["table"]) for term in table_searches
            )
            param_match = (not param_searches) or any(
                re.search(term, search_data["params"]) for term in param_searches
            )
            if glob_match and table_match and param_match:
                search_hits.append(query)
        return sorted(search_hits)

    def query_help(self, query_name: str):
        """
        Print help for `query_name`.

        Parameters
        ----------
        query_name : str
            The name of the query.

        """
        self.query_store[query_name].help()

    def get_query(self, query_name: str) -> str:
        """
        Return the raw query text for `query_name`.

        Parameters
        ----------
        query_name : str
            The name of the query.

        """
        return self.query_store[query_name].query

    def browse_queries(self, **kwargs):
        """
        Return QueryProvider query browser.

        Other Parameters
        ----------------
        kwargs :
            passed to SelectItem constructor.

        Returns
        -------
        SelectItem
            SelectItem browser for TI Data.

        """
        return query_browser()(self, **kwargs)

    # alias for browse_queries
    browse = browse_queries

    def add_custom_query(
        self,
        name: str,
        query: str,
        family: Union[str, Iterable[str]],
        description: Optional[str] = None,
        parameters: Optional[Iterable[QueryParam]] = None,
    ):
        """
        Add a custom function to the provider.

        Parameters
        ----------
        name : str
            The name of the query.
        query : str
            The query text (optionally parameterized).
        family : Union[str, Iterable[str]]
            The query group/family or list of families. The query will
            be added to attributes of the query provider with these
            names.
        description : Optional[str], optional
            Optional description (for query help), by default None
        parameters : Optional[Iterable[QueryParam]], optional
            Optional list of parameter definitions, by default None.
            If the query is parameterized you must supply definitions
            for the parameters here - at least name and type.
            Parameters can be the named tuple QueryParam (also
            exposed as QueryProvider.Param) or a 4-value

        Examples
        --------
        >>> qp = QueryProvider("MSSentinel")
        >>> qp_host = qp.create_param("host_name", "str", "Name of Host")
        >>> qp_start = qp.create_param("start", "datetime")
        >>> qp_end = qp.create_param("end", "datetime")
        >>> qp_evt = qp.create_param("event_id", "int", None, 4688)
        >>>
        >>> query = '''
        >>> SecurityEvent
        >>> | where EventID == {event_id}
        >>> | where TimeGenerated between (datetime({start}) .. datetime({end}))
        >>> | where Computer has "{host_name}"
        >>> '''
        >>>
        >>> qp.add_custom_query(
        >>>     name="test_host_proc",
        >>>     query=query,
        >>>     family="Custom",
        >>>     parameters=[qp_host, qp_start, qp_end, qp_evt]
        >>> )

        """
        if parameters:
            param_dict = {
                param[0]: {
                    "type": param[1],
                    "default": param[2],
                    "description": param[3],
                }
                for param in parameters
            }
        else:
            param_dict = {}
        source = {
            "args": {"query": query},
            "description": description,
            "parameters": param_dict,
        }
        metadata = {"data_families": [family] if isinstance(family, str) else family}
        query_source = QuerySource(
            name=name, source=source, defaults={}, metadata=metadata
        )
        self.query_store.add_data_source(query_source)
        self._add_query_functions()


def _normalize_to_regex(
    search_term: Union[str, Iterable[str], None], ignore_case: bool
) -> List[Pattern[str]]:
    """Return iterable or str search term as list of compiled reg expressions."""
    if not search_term:
        return []
    regex_opts = [re.IGNORECASE] if ignore_case else []
    if isinstance(search_term, str):
        return [re.compile(search_term, *regex_opts)]
    if isinstance(search_term, abc.Iterable):
        return [re.compile(term, *regex_opts) for term in set(search_term)]
    return []



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
from abc import ABC
from enum import IntEnum
from typing import Any, Dict, List, Optional
from weakref import WeakValueDictionary

from IPython.display import display

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_WIDGET_REG: WeakValueDictionary = WeakValueDictionary()


# pylint: disable=too-few-public-methods
class RegisteredWidget(ABC):
    """
    Register widget in the widget registry.

    Registered widgets will store their values in the register.
    Each widget has an ID that that is derived from one or more of the
    initialization parameters. If an instance of the same widget class is
    created with the same parameters, its previous value will be repopulated
    from the registry.
    This is especially useful in notebooks where people accidentally re-run
    the same cell after entering values.
    """

    ALLOWED_KWARGS = ["id_vals", "val_attrs", "nb_params", "name_space", "register"]
    _NB_PARAMS: Dict[str, str] = {}

    def __init__(
        self,
        id_vals: Optional[List[Any]] = None,
        val_attrs: Optional[List[str]] = None,
        nb_params: Optional[Dict[str, str]] = None,
        name_space: Dict[str, Any] = globals(),
        register: bool = True,
        **kwargs,
    ):
        """
        Initialize a registered widget.

        Parameters
        ----------
        id_vals : Optional[List[Any]], optional
            The list of parameter values to use to identify this widget instance,
            by default None
        val_attrs : Optional[List[str]], optional
            The names of the attributes to persist in the registry
            and recall, by default ["value"]
        nb_params : Optional[Dict[str, str]], optional
            A dictionary of attribute names and global variables. If the variable
            exists in the global namespace it will be used to populate the
            corresponding widget attribute. This is only done if the widget
            attribute currently has no value (i.e. restoring a value from
            the registry takes priority over this),
            by default None
        name_space : Dict[str, Any], optional
            Namespace to look for global variables, by default None
        register : bool
            Do not register the widget or retrieve values from previously-
            registered instance.

        """
        del kwargs  # allow to be called with kwargs that are ignored
        # Try to retrieve previous values based on ID of this control
        if register and id_vals:
            id_list = [self.__class__.__name__, *[str(val) for val in id_vals]]
            self._id = hash("".join(id_list))

            if not val_attrs:
                val_attrs = ["value"]
            if self._id in _WIDGET_REG:
                for attr in val_attrs:
                    if hasattr(_WIDGET_REG[self._id], attr):
                        setattr(self, attr, getattr(_WIDGET_REG[self._id], attr))
            # register the current instance as the last instance
            _WIDGET_REG[self._id] = self

        # if there are any notebook params relevant to this control
        if nb_params and name_space:
            for attr, nb_param in nb_params.items():
                # if this doesn't have a value set explicitly or
                # one that was recovered from the widget registry
                # set it from the nb_param value
                wgt_internal_name = self._NB_PARAMS.get(attr, attr)
                if nb_param in name_space and not getattr(
                    self, wgt_internal_name, None
                ):
                    setattr(self, wgt_internal_name, name_space[nb_param])


# pylint: enable=too-few-public-methods


class IPyDisplayMixin:
    """IPython display mixin class."""

    def display(self):
        """Display the interactive widgets."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


class TimeUnit(IntEnum):
    """Time unit enumeration and value."""

    SECOND = 1
    MINUTE = 60
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR
    WEEK = 7 * DAY


def parse_time_unit(unit_str: str) -> TimeUnit:
    """Return the TimeUnit enum matching the input string."""
    if unit_str.casefold().startswith("m"):
        return TimeUnit.MINUTE
    if unit_str.casefold().startswith("h"):
        return TimeUnit.HOUR
    if unit_str.casefold().startswith("s"):
        return TimeUnit.SECOND
    if unit_str.casefold().startswith("d"):
        return TimeUnit.DAY
    if unit_str.casefold().startswith("w"):
        return TimeUnit.WEEK
    return TimeUnit.MINUTE


def default_max_buffer(max_default: Optional[int], default: int, unit: TimeUnit) -> int:
    """Return the max time buffer for a give time unit."""
    mag_default = abs(int(default * 4))
    if max_default is not None:
        max_value = abs(max_default)
        return max(max_value, mag_default)
    if unit == TimeUnit.DAY:
        return max(28, mag_default)
    if unit == TimeUnit.HOUR:
        return max(72, mag_default)
    if unit == TimeUnit.WEEK:
        return max(20, mag_default)
    return max(240, mag_default)


def default_before_after(default: Optional[int], unit: TimeUnit) -> int:
    """Return default before and after bounds for a TimeUnit."""
    if default is not None:
        return abs(default)
    if unit in (TimeUnit.DAY, TimeUnit.WEEK):
        return 1
    if unit == TimeUnit.HOUR:
        return 6
    return 60



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""VirusTotal File Behavior functions."""
import re
from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, List, Optional, Union

import attr
import ipywidgets as widgets
import numpy as np
import pandas as pd

from ..._version import VERSION
from ...common.exceptions import MsticpyImportExtraError, MsticpyUserError
from ...transform.proc_tree_builder import ProcSchema, build_proc_tree

try:
    import vt
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without vt-py and vt-graph-api packages installed.",
        title="Error importing VirusTotal modules.",
        extra="vt3",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"


VT_API_NOT_FOUND = "NotFoundError"


_FB_CAT_PATTERNS = {
    "File": "file.*",
    "Process": "process.*|command.*|module.*",
    "Registry": "registry.*",
    "Network": ".*ips|dns.*|.*urls|ip.*|http.*|tls",
    "System": "mutex.*|calls.*|permissions.*|text.*",
    "Other": ".*",
}

_BORDER_LAYOUT = widgets.Layout(
    **{
        "width": "90%",
        "border": "solid gray 1px",
        "margin": "1pt",
        "padding": "5pt",
    }
)


class VTFileBehavior:
    """VirusTotal File Behavior class."""

    _SANDBOXES = [
        "f_secure_sandbox",
        "bitdam_atp",
        "vmray",
        "virustotal_zenbox",
        "sangfor_zsand",
        "virustotal_jujubox",
        "qianxin_reddrip",
        "nsfocus_poma",
        "virustotal_androbox",
        "venuseye_sandbox",
        "cyber_adapt",
        "dr_web_vxcube",
        "virustotal_observer",
        "tencent_habo",
        "yomi_hunter",
        "virustotal_jsbox",
        "virustotal_cuckoofork",
        "lastline",
        "reaqta_hive",
        "os_x_sandbox",
        "virustotal_droidy",
        "sndbox",
        "virustotal_r2dbox",
        "rising_moves",
        "virustotal_box_of_apples",
        "secondwrite",
        "malwation",
        "c2ae",
        "microsoft_sysinternals",
    ]

    _FP_ENDPOINTS = {
        "summary": "/files/{id}/behaviour_summary",
        "sandbox": "/file_behaviours/{id}_{sandbox}",
        "evtx": "/file_behaviours/{sandbox_id}/evtx",
        "pcap": "/file_behaviours/{sandbox_id}/pcap",
        "memdump": "/file_behaviours/{sandbox_id}/memdump",
    }

    @classmethod
    def list_sandboxes(cls) -> List[str]:
        """Return list of known sandbox types."""
        return list(cls._SANDBOXES)

    def __init__(
        self,
        vt_key: str = None,
        file_id: Optional[str] = None,
        file_summary: Optional[Union[pd.DataFrame, pd.Series, Dict[str, Any]]] = None,
    ):
        """
        Initialize the VTFileBehavior class.

        Parameters
        ----------
        vt_key : str, optional
            VirusTotal API key, by default None
        file_id : Optional[str], optional
            The ID of the file to look up, by default None
        file_summary : Optional[Union[pd.DataFrame, pd, Series, Dict[str, Any]]], optional
            VT file summary - this can be in one of the following formats:
            VT object dictionary
            Pandas DataFrame - first row is assumed to be the file summary
            Pandas Series
            by default None

        """
        self._vt_client = vt.Client(apikey=vt_key)
        if file_id is None and file_summary is None:
            raise MsticpyUserError(
                "You must supply either a file_id or a file_summary.",
                title="Missing required parameter.",
            )

        if isinstance(file_summary, pd.DataFrame):
            file_summary = file_summary.iloc[0]
        if isinstance(file_summary, pd.Series):
            file_summary = file_summary.to_dict()
        self.file_summary = file_summary or {}  # type: ignore
        self.file_id = file_id or self.file_summary.get("id")

        self._file_behavior: Dict[str, Any] = {}
        self.categories: Dict[str, Any] = {}
        self.behavior_links: Dict[str, Any] = {}
        self.process_tree_df: Optional[pd.DataFrame] = None

    def _reset_summary(self):
        self._file_behavior: Dict[str, Any] = {}
        self.categories: Dict[str, Any] = {}
        self.process_tree_df: Optional[pd.DataFrame] = None

    @property
    def sandbox_id(self) -> str:
        """Return sandbox ID of detonation."""
        return self.categories.get("id", "")

    @property
    def has_evtx(self) -> bool:
        """Return True if EVTX data is available (Enterprise only)."""
        return self.categories.get("has_evtx", False)

    @property
    def has_memdump(self) -> bool:
        """Return True if memory dump data is available (Enterprise only)."""
        return self.categories.get("has_memdump", False)

    @property
    def has_pcap(self) -> bool:
        """Return True if PCAP data is available (Enterprise only)."""
        return self.categories.get("has_pcap", False)

    def get_file_behavior(self, sandbox: str = None):
        """
        Retrieve the file behavior data.

        Parameters
        ----------
        sandbox : str, optional
            Name of specific sandbox to retrieve, by default None
            If None, it will retrieve the behavior summary.

        """
        if sandbox:
            endpoint = self._FP_ENDPOINTS["sandbox"].format(
                id=self.file_id,
                sandbox=sandbox,
            )
        else:
            endpoint = self._FP_ENDPOINTS["summary"].format(id=self.file_id)

        try:
            self._file_behavior = self._vt_client.get_data(endpoint)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                self._file_behavior = {"id": self.file_id, "result": VT_API_NOT_FOUND}
                return
            raise
        finally:
            self._vt_client.close()

        if "attributes" in self._file_behavior:
            self.categories = self._file_behavior.get("attributes", {})
            self.behavior_links = self._file_behavior.get("links", {})
        else:
            self.categories = self._file_behavior

    def browse(self) -> Optional[widgets.VBox]:
        """Browse the behavior categories."""
        if not self.has_behavior_data:
            self._print_no_data()
            return None
        groupings = {}
        remaining_categories = set(self.categories)
        for name, pattern in _FB_CAT_PATTERNS.items():
            groupings[name] = _extract_subcats(pattern, remaining_categories)
            remaining_categories = remaining_categories - groupings[name]

        accordion = widgets.Accordion()
        child_tabs = {}
        for group, sub_cats in groupings.items():
            sub_cat_tab = widgets.Tab()
            tab_content = {
                section: widgets.HTML(value=_format_widget_data(items))
                for section, items in self.categories.items()
                if items and section in sub_cats
            }
            sub_cat_tab.children = list(tab_content.values())
            for idx, section in enumerate(tab_content):
                sub_cat_tab.set_title(idx, section)
            child_tabs[group] = sub_cat_tab

        accordion.children = list(child_tabs.values())
        for idx, group_name in enumerate(child_tabs):
            accordion.set_title(idx, group_name)
        accordion.selected_index = 0

        html_title = widgets.HTML(
            ("<h2>VirusTotal Detonation Details</h2>" f"For file {self.file_id}"),
            layout=_BORDER_LAYOUT,
        )
        return widgets.VBox([html_title, accordion])

    @property
    def process_tree(self) -> Any:
        """Return the process tree plot."""
        if not self.has_behavior_data:
            self._print_no_data()
            return None
        if self.process_tree_df is None:
            self.process_tree_df = _build_process_tree(self.categories)
        plot, _ = self.process_tree_df.mp_plot.process_tree(
            schema=VT_PROCSCHEMA,
            legend_col="name",
            hide_legend=True,
        )
        return plot

    @property
    def has_behavior_data(self) -> bool:
        """Return true if file behavior data available."""
        return bool(self.categories)

    def _print_no_data(self):
        """Print a message if operation is tried with no data."""
        print(f"No data available for {self.file_id}.")


# Process tree extraction


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True)
class SIProcess:
    """Data class to hold each process from detonation."""

    process_id: str
    name: str
    cmd_line: str
    parent_id: int = -1
    proc_key: Optional[str] = None
    parent_key: Optional[str] = None
    path: Optional[str] = None
    IsRoot: bool = False
    IsLeaf: bool = False
    IsBranch: bool = False
    children: list = []
    # proc_children: list = []
    time_offset: int = 0


# pylint: enable=too-few-public-methods


VT_PROCSCHEMA = ProcSchema(
    **{
        "process_name": "name",
        "process_id": "process_id",
        "parent_id": "parent_id",
        "cmd_line": "cmd_line",
        "time_stamp": "time_stamp",
        "logon_id": "logon_id",
        "path_separator": "\\",
        "user_name": "user_name",
        "host_name_column": "host",
        "event_id_column": "event_id",
    }
)


def _build_process_tree(fb_categories):
    """Top level function to create displayable DataFrame."""
    proc_tree_raw = deepcopy(fb_categories["processes_tree"])
    procs_created = {
        Path(proc).parts[-1].lower(): proc
        for proc in fb_categories["processes_created"]
    }

    si_procs = _extract_processes(proc_tree_raw, procs_created)
    process_tree_df = pd.DataFrame(_procs_to_df(si_procs)).drop(columns="children")
    process_tree_df = _try_match_commandlines(
        fb_categories["command_executions"], process_tree_df
    )
    return _fill_missing_proc_tree_values(process_tree_df)


def _extract_processes(process_data, procs_created, parent=None):
    """Convert processes_tree attribute to SIProcessObjects."""
    procs = []
    for process in process_data:
        si_proc = _create_si_proc(process, procs_created)
        # pylint: disable=invalid-name
        if parent:
            si_proc.parent_key = parent.proc_key
            si_proc.IsBranch = True
        else:
            si_proc.IsRoot = True
        child_procs_raw = process.get("children", [])
        if child_procs_raw:
            si_proc.children = _extract_processes(
                child_procs_raw, procs_created, parent=si_proc
            )
        else:
            si_proc.IsLeaf = True
            si_proc.IsBranch = False
        procs.append(si_proc)
    return procs


def _create_si_proc(raw_proc, procs_created):
    """Return an SIProcess Object from a raw VT proc definition."""
    # raw_proc = copy(raw_proc)
    name = raw_proc.get("name")
    raw_proc["cmd_line"] = name
    for proc in procs_created:
        if name.lower().endswith(proc):
            raw_proc["name"] = procs_created[proc]
            break
    raw_proc["proc_key"] = raw_proc["process_id"] + "|" + raw_proc["name"]
    # print(name, raw_proc.keys())
    return SIProcess(**raw_proc)


# Convert to DF
def _procs_to_df(procs):
    """Convert the SIProcess objects recursively to a list."""
    df_list = []
    for proc in procs:
        df_list.append(attr.asdict(proc))
        if proc.children:
            df_list.extend(_procs_to_df(proc.children))
    return df_list


# Try to Match up 'command_executions' commandline data with
# process_df.
def _try_match_commandlines(
    command_executions, procs_cmds: pd.DataFrame
) -> pd.DataFrame:
    """Return DF with matched commandlines."""
    procs_cmd = procs_cmds.copy()
    procs_cmd["cmd_line"] = np.nan
    weak_matches = 0
    for cmd in command_executions:
        for idx, row in procs_cmd.iterrows():
            # print(row["name"], cmd, row["cmd_line"], isinstance(row["cmd_line"], str))
            if (
                not isinstance(row["cmd_line"], str)
                and np.isnan(row["cmd_line"])
                and row["name"] in cmd
            ):
                # print("Found match:", row["name"], "==", cmd)
                procs_cmd.loc[idx, "cmd_line"] = cmd  # type: ignore
                break
    for cmd in command_executions:
        for idx, row in procs_cmd.iterrows():
            # print(row["name"], cmd, row["cmd_line"], isinstance(row["cmd_line"], str))
            if (
                not isinstance(row["cmd_line"], str)
                and np.isnan(row["cmd_line"])
                and Path(row["name"]).stem.lower() in cmd.lower()
            ):
                weak_matches += 1
                # print("Found weak match:", row["name"], "~=", cmd)
                procs_cmd.loc[idx, "cmd_line"] = cmd  # type: ignore
                break

    if weak_matches:
        print(
            f"WARNING: {weak_matches} of the {len(command_executions)} commandlines",
            "were weakly matched - some commandlines may be attributed",
            "to the wrong instance of the process.",
            end="\n",
        )
    return procs_cmd


def _fill_missing_proc_tree_values(process_df: pd.DataFrame) -> pd.DataFrame:
    # Define a schema to map Df names on to internal ProcSchema
    process_df["path"] = np.nan
    process_df.loc[process_df.IsRoot, "path"] = pd.Series(
        process_df[process_df.IsRoot].index.astype("str")
    ).apply(lambda x: x.zfill(5))

    # Fill in some required fields with placeholder data
    process_df["time_stamp"] = datetime.now(tz=timezone.utc)
    process_df["host"] = "sandbox"
    process_df["logon_id"] = "na"
    process_df["event_id"] = "na"
    process_df["source_index"] = pd.Series(process_df.index.astype("str")).apply(
        lambda x: x.zfill(5)
    )

    proc_tree = process_df.set_index("proc_key")

    first_unique = proc_tree.index.duplicated()
    proc_tree = proc_tree[~first_unique]
    # msticpy function to build the tree
    return build_proc_tree(proc_tree)


# Process browser helper functions
def _extract_subcats(pattern, categs):
    """Extract the category names matching `pattern`."""
    return {cat for cat in categs if re.match(pattern, cat)}


def _format_widget_data(data_item):
    if not data_item:
        return ""
    if isinstance(data_item, list):
        if isinstance(data_item[0], dict):
            return pd.DataFrame(data_item).style.hide(axis="index").to_html()
        if isinstance(data_item[0], str):
            return pd.DataFrame(pd.Series(data_item)).style.hide(axis="index").to_html()
    return f"<pre>{pformat(data_item)}</pre>"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider sub-package."""

from ...lazy_importer import lazy_import

_LAZY_IMPORTS = {
    "msticpy.context.azure.azure_data.AzureData",
    "msticpy.context.azure.sentinel_core.MicrosoftSentinel",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP Context Provider base.

HTTP implementation of a ContextProvider.
It is used to interface with HTTP API providing additional contexts.
It inherits from ContextProvider and HttpProvider

"""
from functools import lru_cache
from json import JSONDecodeError
from typing import Any, Dict

import pandas as pd

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ..http_provider import APILookupParams, HttpProvider
from ..lookup_result import LookupStatus
from .context_provider_base import ContextProvider

__version__ = VERSION
__author__ = "Florian Bracq"


# Create alias for simplicity
ContextLookupParams = APILookupParams


class HttpContextProvider(ContextProvider, HttpProvider):
    """HTTP Context Provider base class."""

    @lru_cache(maxsize=256)
    def lookup_observable(
        self,
        observable: str,
        observable_type: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        observable : str
            observable to lookup
        observable_type : str, optional
            The Type of the observable to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """
        result = self._check_observable_type(
            observable, observable_type, query_subtype=query_type
        )

        result["Provider"] = kwargs.get("provider_name", self.__class__.__name__)

        req_params: Dict[str, Any] = {}
        try:
            verb, req_params = self._substitute_parms(
                result["SafeObservable"], result["ObservableType"], query_type
            )
            if verb == "GET":
                response = self._httpx_client.get(
                    **req_params, timeout=get_http_timeout(**kwargs)
                )
            else:
                raise NotImplementedError(f"Unsupported verb {verb}")

            result["Status"] = response.status_code
            result["Reference"] = req_params["url"]
            if result["Status"] == 200:
                try:
                    result["RawResult"] = response.json()
                    result["Result"], result["Details"] = self.parse_results(result)
                except JSONDecodeError:
                    result[
                        "RawResult"
                    ] = f"""There was a problem parsing results from this lookup:
                                        {response.text}"""
                    result["Result"] = False
                    result["Details"] = {}
                result["Status"] = LookupStatus.OK.value
            else:
                result["RawResult"] = str(response.text)
                result["Result"] = False
                result["Details"] = self._response_message(result["Status"])
        except (
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            self._err_to_results(result, err)
            if not isinstance(err, LookupError):
                url = req_params.get("url", None) if req_params else None
                result["Reference"] = url
        return pd.DataFrame([result])



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree Visualization."""
import warnings
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

import numpy as np
import pandas as pd
from bokeh.io import output_notebook, reset_output, show

# pylint: enable=no-name-in-module
try:
    from bokeh.core.property.vectorization import Field
except ImportError:
    Field = dict  # type: ignore
from bokeh.layouts import column, row
from bokeh.models import (  # type: ignore[attr-defined]
    BoxSelectTool,
    ColorBar,
    ColumnDataSource,
    CustomJS,
    GestureTool,
    HoverTool,
    LayoutDOM,
    RangeTool,
)
from bokeh.models.widgets import DataTable, DateFormatter, TableColumn

# pylint: disable=no-name-in-module
from bokeh.palettes import viridis
from bokeh.plotting import figure
from bokeh.transform import dodge, factor_cmap, linear_cmap
from deprecated.sphinx import deprecated

from .._version import VERSION
from ..common.utility import check_kwargs, export
from ..transform.proc_tree_builder import build_process_tree, infer_schema
from ..transform.proc_tree_schema import ColNames as Col
from ..transform.proc_tree_schema import ProcessTreeSchemaException, ProcSchema

# pylint: disable=unused-import
# flake8: noqa F401
from ..transform.process_tree_utils import (
    get_ancestors,
    get_children,
    get_descendents,
    get_parent,
    get_process,
    get_process_key,
    get_root,
    get_root_tree,
    get_roots,
    get_siblings,
    get_summary_info,
    get_tree_depth,
)
from .figure_dimension import bokeh_figure

__all__ = [
    "get_ancestors",
    "get_children",
    "get_descendents",
    "get_parent",
    "get_process",
    "get_process_key",
    "get_root",
    "get_root_tree",
    "get_roots",
    "get_siblings",
    "get_summary_info",
    "get_tree_depth",
]

# pylint: enable=unused-import

__version__ = VERSION
__author__ = "Ian Hellen"

_DEFAULT_KWARGS = ["height", "title", "width", "hide_legend", "pid_fmt"]

# wrap figure function to handle v2/v3 parameter renaming
figure = bokeh_figure(figure)  # type: ignore[assignment, misc]


@export
def build_and_show_process_tree(
    data: pd.DataFrame,
    schema: Union[ProcSchema, Dict[str, Any]] = None,
    output_var: str = None,
    legend_col: str = None,
    **kwargs,
) -> Tuple[figure, LayoutDOM]:
    """
    Build process tree from data and plot a tree.

    Parameters
    ----------
    data : pd.DataFrame
        Window process creation or Linux Auditd events
    schema : Union[ProcSchema, Dict[str, Any]], optional
        The column schema to use, by default None.
        If supplied as a dict it must include definitions for the
        required fields in the ProcSchema class
        If None, then the schema is inferred
    output_var : str, optional
        Output variable for selected items in the tree,
        by default None
    legend_col : str, optional
        The column used to color the tree items, by default None
    kwargs : Dict[str, Any]
        Additional arguments passed to plot_process_tree

    Other Parameters
    ----------------
    height : int, optional
        The height of the plot figure
        (the default is 700)
    width : int, optional
        The width of the plot figure (the default is 900)
    title : str, optional
        Title to display (the default is None)
    hide_legend : bool, optional
        Hide the legend box, even if legend_col is specified.
    pid_fmt : str, optional
        Display Process ID as 'dec' (decimal), 'hex' (hexadecimal),
        or 'guid' (string), default is 'hex'.

    Returns
    -------
    Tuple[figure, LayoutDOM]:
        figure - The main bokeh.plotting.figure
        Layout - Bokeh layout structure.

    Notes
    -----
    For full parameter set for process tree display see the
    help for plot_process_tree.

    See Also
    --------
    plot_process_tree

    """
    if isinstance(schema, dict):
        schema = ProcSchema(**schema)
    # Check if this table already seems to have the proc_tree metadata
    missing_cols = _check_proc_tree_schema(data)
    plot_args = {**kwargs, "legend_col": legend_col}
    if missing_cols:
        data = build_process_tree(procs=data, schema=schema, plot_args=plot_args)
    return plot_process_tree(data, schema, output_var=output_var, **plot_args)


# pylint: disable=too-many-locals, too-many-statements
@export
def plot_process_tree(  # noqa: MC0001
    data: pd.DataFrame,
    schema: Union[ProcSchema, Dict[str, Any]] = None,
    output_var: str = None,
    legend_col: str = None,
    show_table: bool = False,
    **kwargs,
) -> Tuple[figure, LayoutDOM]:
    """
    Plot a Process Tree Visualization.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing one or more Process Trees
    schema : Union[ProcSchema, Dict[str, Any]], optional
        The column schema to use, by default None.
        If supplied as a dict it must include definitions for the
        required fields in the ProcSchema class
        If None, then the schema is inferred
    output_var : str, optional
        Output variable for selected items in the tree,
        by default None
    legend_col : str, optional
        The column used to color the tree items, by default None
    show_table: bool
        Set to True to show a data table, by default False.

    Other Parameters
    ----------------
    height : int, optional
        The height of the plot figure
        (the default is 700)
    width : int, optional
        The width of the plot figure (the default is 900)
    title : str, optional
        Title to display (the default is None)
    hide_legend : bool, optional
        Hide the legend box, even if legend_col is specified.
    pid_fmt : str, optional
        Display Process ID as 'dec' (decimal), 'hex' (hexadecimal),
        or 'guid' (string), default is 'hex'.

    Returns
    -------
    Tuple[figure, LayoutDOM]:
        figure - The main bokeh.plotting.figure
        Layout - Bokeh layout structure.

    Raises
    ------
    ProcessTreeSchemaException
        If the data set schema is not valid for the plot.

    Notes
    -----
    The `output_var` variable will be overwritten with any selected
    values.

    """
    check_kwargs(kwargs, _DEFAULT_KWARGS)
    reset_output()
    output_notebook()

    plot_height: int = kwargs.pop("height", 700)
    plot_width: int = kwargs.pop("width", 900)
    title: str = kwargs.pop("title", "ProcessTree")
    hide_legend = kwargs.pop("hide_legend", False)
    pid_fmt = kwargs.pop("pid_fmt", "hex")

    proc_data, schema, levels, n_rows = _pre_process_tree(data, schema, pid_fmt=pid_fmt)
    if schema is None:
        raise ProcessTreeSchemaException("Could not infer data schema from data set.")
    if levels is None:
        raise ProcessTreeSchemaException("Could create process relationships.")

    source = ColumnDataSource(data=proc_data)
    # Get legend/color bar map
    fill_map, color_bar = _create_fill_map(source, legend_col)

    max_level = max(levels) + 3
    min_level = min(levels)

    if color_bar:
        title += " (color bar = {legend_col})"
    visible_range = int(plot_height / 35)
    y_start_range = (n_rows - visible_range, n_rows + 1)
    b_plot = figure(
        title=title,
        x_range=(min_level, max_level),
        y_range=y_start_range,
        tools=["reset", "save", "tap", "ywheel_pan"],
        toolbar_location="above",
        active_scroll="ywheel_pan",
        width=plot_width,
        height=plot_height,
    )

    hover = HoverTool(
        tooltips=_get_tool_tips(schema),
        formatters={f"@{schema.time_stamp}": "datetime"},
    )
    b_plot.add_tools(hover)

    # dodge to align rectangle with grid
    rect_x = dodge("Level", 1.75, range=b_plot.x_range)  # type: ignore
    rect_plot_params = {
        "width": 3.5,
        "height": 0.95,
        "source": source,
        "fill_alpha": 0.4,
        "fill_color": fill_map,
    }

    if color_bar:
        b_plot.add_layout(color_bar, "right")
    elif legend_col:
        rect_plot_params["legend_field"] = legend_col
    rect_plot = b_plot.rect(x=rect_x, y="Row", **rect_plot_params)
    if legend_col and not color_bar:
        b_plot.legend.title = legend_col
        b_plot.legend.label_text_font_size = "7pt"
    if hide_legend:
        b_plot.legend.visible = False

    text_props = {"source": source, "text_align": "left", "text_baseline": "middle"}

    def x_dodge(x_offset):
        return dodge("Level", x_offset, range=b_plot.x_range)

    def y_dodge(y_offset):
        return dodge("Row", y_offset, range=b_plot.y_range)

    b_plot.text(
        x=x_dodge(0.1),
        y=y_dodge(-0.2),
        text="__cmd_line$$",
        text_font_size="7pt",
        **text_props,
    )
    b_plot.text(
        x=x_dodge(0.1),
        y=y_dodge(0.25),
        text="__proc_name$$",
        text_font_size="8pt",
        **text_props,
    )
    b_plot.text(
        x=x_dodge(2.2),
        y=y_dodge(0.25),
        text="__proc_id$$",
        text_font_size="8pt",
        **text_props,
    )

    # Plot options
    _set_plot_option_defaults(b_plot)
    b_plot.xaxis.ticker = sorted(levels)
    b_plot.xgrid.ticker = sorted(levels)
    b_plot.hover.renderers = [rect_plot]  # only hover element boxes

    # Selection callback
    if output_var is not None:
        get_selected = _create_js_callback(source, output_var)
        b_plot.js_on_event("tap", get_selected)
        box_select = BoxSelectTool(callback=get_selected)
        b_plot.add_tools(box_select)

    range_tool = _create_vert_range_tool(
        data=source,
        min_y=0,
        max_y=n_rows,
        plot_range=b_plot.y_range,
        width=90,
        height=plot_height,
        x_col="Level",
        y_col="Row",
        fill_map=fill_map,
    )
    plot_elems: LayoutDOM = row(b_plot, range_tool)  # type: ignore
    if show_table:
        data_table = _create_data_table(source, schema, legend_col)
        plot_elems = column(plot_elems, data_table)
    show(plot_elems)  # type: ignore
    return b_plot, plot_elems


# pylint: enable=too-many-locals, too-many-statements


class TreeResult(NamedTuple):
    """Result tuple for _pre_process_tree."""

    proc_tree: pd.DataFrame
    schema: Optional[ProcSchema]
    levels: Optional[np.ndarray]
    n_rows: int


def _pre_process_tree(
    proc_tree: pd.DataFrame,
    schema: Union[Dict[str, Any], ProcSchema] = None,
    pid_fmt: str = "hex",
) -> TreeResult:
    """Extract dimensions and formatted values from proc_tree."""
    # Check if this table already seems to have the proc_tree metadata
    missing_cols = _check_proc_tree_schema(proc_tree)

    if isinstance(schema, dict):
        schema = ProcSchema(**schema)
    if missing_cols:
        proc_tree = build_process_tree(procs=proc_tree, schema=schema)

    if schema is None:
        schema = infer_schema(proc_tree)
    if schema is None:
        return TreeResult(proc_tree, None, None, 0)

    _validate_plot_schema(proc_tree, schema)

    # kludgy fix to prevent NaNs making it into the data - Bokeh 3.0
    # is very sensitive to this in some places.
    proc_tree = proc_tree.fillna("NA")
    proc_tree = proc_tree.sort_values(
        by=["path", schema.time_stamp], ascending=True
    ).reset_index()
    n_rows = len(proc_tree)
    proc_tree["Row"] = proc_tree.index
    proc_tree["Row"] = n_rows - proc_tree["Row"]
    proc_tree["Level"] = proc_tree["path"].str.count("/") + 1

    levels = proc_tree["Level"].unique()

    proc_tree[schema.process_name] = proc_tree[schema.process_name].fillna("unknown")
    proc_tree["__proc_name$$"] = proc_tree.apply(  # type: ignore
        lambda x: x[schema.process_name].split(schema.path_separator)[-1], axis=1
    )
    proc_tree[schema.process_id] = proc_tree[schema.process_id].fillna("unknown")
    proc_tree["__proc_id$$"] = proc_tree[schema.process_id].apply(
        _pid_fmt, args=(pid_fmt,)
    )

    # Command line processing
    if not schema.cmd_line:
        schema = ProcSchema(**vars(schema))
        schema.cmd_line = "__cmd_line__"
        proc_tree[schema.cmd_line] = "cmdline unknown"
    # trim long command lines
    max_cmd_len = 500 // len(levels)
    proc_tree[schema.cmd_line] = proc_tree[schema.cmd_line].astype(str)
    long_cmd = proc_tree[schema.cmd_line].str.len() > max_cmd_len
    proc_tree.loc[long_cmd, "__cmd_line$$"] = (
        proc_tree[long_cmd][schema.cmd_line].str[:max_cmd_len] + "..."
    )
    # replace missing cmd lines
    proc_tree.loc[~long_cmd, "__cmd_line$$"] = proc_tree[~long_cmd][
        schema.cmd_line
    ].fillna("cmdline unknown")
    return TreeResult(proc_tree=proc_tree, schema=schema, levels=levels, n_rows=n_rows)


def _pid_fmt(pid, pid_fmt):
    """Format process ID in required string format."""
    if pid == np.nan:
        pid = ""
    if pid_fmt == "hex":
        return (
            f"PID: {pid}"
            if str(pid).startswith("0x")
            else f"PID: 0x{int(pid):x}"
            if isinstance(pid, int)
            else "NA"
        )
    if pid_fmt == "guid":
        return f"GUID: {pid}"
    return (
        f"PID: {pid}"
        if not str(pid).startswith("0x")
        else f"PID: {int(pid, base=16)}"
        if isinstance(pid, int)
        else "NA"
    )


def _validate_plot_schema(proc_tree: pd.DataFrame, schema):
    """Validate that we have the required columns."""
    required_cols = {"path", schema.process_name, schema.process_id}
    proc_cols = set(proc_tree.columns)
    missing = required_cols - proc_cols
    if missing:
        raise ProcessTreeSchemaException(
            f"Required columns not found in data set: {','.join(missing)}"
        )


def _set_plot_option_defaults(b_plot):
    """Set default plot options."""
    b_plot.outline_line_color = None
    b_plot.grid.grid_line_color = "navy"
    b_plot.axis.axis_line_color = None
    b_plot.axis.major_tick_line_color = "navy"
    b_plot.xaxis.visible = False
    b_plot.yaxis.visible = False
    b_plot.xgrid.visible = True
    b_plot.ygrid.visible = False
    b_plot.xgrid.minor_grid_line_color = "navy"
    b_plot.xgrid.minor_grid_line_alpha = 0.1
    b_plot.xgrid.grid_line_color = "navy"
    b_plot.xgrid.grid_line_alpha = 0.1
    b_plot.axis.major_label_standoff = 0


def _get_tool_tips(schema: ProcSchema):
    """Return tool tip formatter."""
    return [
        ("Process", f"@{schema.process_name}"),
        ("PID", f"@{schema.process_id}"),
        ("CmdLine", f"@{schema.cmd_line}"),
        ("SubjUser", f"@{schema.user_name}"),
        ("SubjLgnId", f"@{schema.logon_id}"),
        ("TgtLgnId", f"@{schema.target_logon_id}"),
        ("Time", f"@{schema.time_stamp}{{%F %T.%3N}}"),
    ]


def _create_js_callback(source: ColumnDataSource, result_var: str) -> CustomJS:
    """Create and return CustomJS callback to set Python variable."""
    ret_var_js = """
        // get data source from Callback args
        var inds = source.selected.indices;
        var output = [];
        for (var i = 0; i < inds.length; i++) {
            output.push(source.data[itemkey][inds[i]])
        }
        var out_str = JSON.stringify(output);
        var py_str = `${output_var} = ${out_str}`;
        console.log(py_str);
        IPython.notebook.kernel.execute(py_str);
    """
    return CustomJS(
        args={"source": source, "itemkey": Col.proc_key, "output_var": result_var},
        code=ret_var_js,
    )


def _create_fill_map(
    source: ColumnDataSource, source_column: str = None
) -> Tuple[Union[Field, str], Optional[ColorBar]]:
    """Create factor map or linear map based on `source_column`."""
    fill_map: Union[str, Field] = "navy"
    color_bar = None
    key_column = source_column or "Level"

    col_kind = source.data[key_column].dtype.kind  # type: ignore[union-attr]
    if col_kind in ["b", "O"]:
        s_values = set(source.data[key_column])
        if np.nan in s_values:
            s_values.remove(np.nan)
        values = list(s_values)
        fill_map = factor_cmap(
            key_column, palette=viridis(max(3, len(values))), factors=values
        )
    elif col_kind in ["i", "u", "f", "M"]:
        values = [val for val in source.data[key_column] if not np.isnan(val)]  # type: ignore
        fill_map = linear_cmap(
            field_name=key_column,
            palette=viridis(256),
            low=np.min(values),
            high=np.max(values),
        )
        if source_column is not None:
            # If user hasn't specified a legend column - don't create a bar
            color_bar = ColorBar(
                color_mapper=fill_map["transform"], width=8, location=(0, 0)  # type: ignore
            )
    return fill_map, color_bar


# pylint: disable=too-many-arguments
def _create_vert_range_tool(
    data, min_y, max_y, plot_range, width, height, x_col, y_col, fill_map="navy"
):
    """Return vertical range too for plot."""
    rng_select = figure(
        y_range=(min_y - 1, max_y + 1),
        toolbar_location=None,
        width=width,
        height=height,
    )

    x_dodge = dodge(x_col, -0.5)
    rng_select.rect(
        x=x_dodge,
        y=y_col,
        width=1.2,
        height=0.8,
        source=data,
        fill_alpha=0.6,
        fill_color=fill_map,
    )

    rng_select.xaxis.visible = False
    rng_select.yaxis.visible = False

    range_tool = RangeTool(y_range=plot_range)
    range_tool.overlay.fill_color = "navy"
    range_tool.overlay.fill_alpha = 0.2
    rng_select.ygrid.grid_line_color = None
    rng_select.xgrid.grid_line_color = None
    rng_select.add_tools(range_tool)
    if isinstance(range_tool, GestureTool):
        rng_select.toolbar.active_multi = range_tool
    return rng_select


# pylint: enable=too-many-arguments


def _create_data_table(
    source: ColumnDataSource, schema: ProcSchema, legend_col: str = None
):
    """Return DataTable widget for source."""
    column_names = [
        schema.user_name,
        schema.user_id,
        schema.logon_id,
        schema.process_id,
        schema.process_name,
        schema.cmd_line,
        schema.parent_id,
        schema.parent_name,
        schema.target_logon_id,
    ]

    if legend_col and legend_col not in column_names:
        column_names.append(legend_col)

    date_fmt = "%F %T"
    columns = [
        TableColumn(
            field=schema.time_stamp,
            title=schema.time_stamp,
            formatter=DateFormatter(format=date_fmt),
        )
    ]
    columns2 = [
        TableColumn(field=col, title=col)
        for col in column_names
        if col in source.column_names
    ]

    return DataTable(source=source, columns=columns + columns2, width=950, height=150)


def _check_proc_tree_schema(data):
    """Return true if expected process tree columns are present."""
    if data.index.name != Col.proc_key:
        return {Col.proc_key}
    expected_cols = {Col.parent_key, "IsRoot", "IsLeaf", "IsBranch", "path"}
    return expected_cols - set(data.columns)


# pylint: disable=too-few-public-methods
@deprecated("Will be removed in version 2.0.0", version="1.7.0")
@pd.api.extensions.register_dataframe_accessor("mp_process_tree")
class ProcessTreeAccessor:
    """Pandas api extension for Process Tree."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def plot(self, **kwargs) -> Tuple[figure, LayoutDOM]:
        """
        Build and plot a process tree.

        Parameters
        ----------
        schema : ProcSchema, optional
            The data schema to use for the data set, by default None
            (if None the schema is inferred)
        output_var : str, optional
            Output variable for selected items in the tree,
            by default None
        legend_col : str, optional
            The column used to color the tree items, by default None
        show_table: bool
            Set to True to show a data table, by default False.

        Other Parameters
        ----------------
        height : int, optional
            The height of the plot figure
            (the default is 700)
        width : int, optional
            The width of the plot figure (the default is 900)
        title : str, optional
            Title to display (the default is None)
        hide_legend : bool, optional
            Hide the legend box, even if legend_col is specified.
        pid_fmt : str, optional
            Display Process ID as 'dec' (decimal), 'hex' (hexadecimal),
            or 'guid' (string), default is 'hex'.

        Returns
        -------
        Tuple[figure, LayoutDOM]:
            figure - The main bokeh.plotting.figure
            Layout - Bokeh layout structure.

        """
        warn_message = (
            "This accessor method has been deprecated.\n"
            "Please use df.mp_plot.process_tree() method instead."
            "This will be removed in MSTICPy v2.2.0"
        )
        warnings.warn(warn_message, category=DeprecationWarning)
        return build_and_show_process_tree(data=self._df, **kwargs)

    def build(self, schema: ProcSchema = None, **kwargs) -> pd.DataFrame:
        """
        Build process trees from the process events.

        Parameters
        ----------
        procs : pd.DataFrame
            Process events (Windows 4688 or Linux Auditd)
        schema : ProcSchema, optional
            The column schema to use, by default None
            If None, then the schema is inferred
        show_summary : bool
            Shows summary of the built tree, default is False. : bool
        debug : bool
            If True produces extra debugging output,
            by default False

        Returns
        -------
        pd.DataFrame
            Process tree dataframe.

        Notes
        -----
        It is not necessary to call this before `plot`. The process
        tree is built automatically. This is only needed if you want
        to return the processed tree data as a DataFrame

        """
        warn_message = (
            "This accessor method has been deprecated.\n"
            "Please use df.mp.build_process_tree() method instead."
            "This will be removed in MSTICPy v2.2.0"
        )
        warnings.warn(warn_message, category=DeprecationWarning)
        return build_process_tree(
            procs=self._df,
            schema=schema,
            show_summary=kwargs.get("show_summary", kwargs.get("show_progress", False)),
            debug=kwargs.get("debug", False),
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP Lookup base class.

Input can be a single item or a pandas DataFrame containing
multiple items. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import traceback
from abc import abstractmethod
from typing import Any, Dict, List, Tuple, Union

import attr
import httpx
import pandas as pd
from attr import Factory

from .._version import VERSION
from ..common.exceptions import MsticpyConfigError
from ..common.pkg_config import get_http_timeout
from ..common.utility import mp_ua_header
from .lookup_result import LookupStatus
from .provider_base import Provider

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
@attr.s(auto_attribs=True)
class APILookupParams:
    """HTTP Lookup Params definition."""

    path: str = ""
    verb: str = "GET"
    full_url: bool = False
    headers: Dict[str, str] = Factory(dict)
    params: Dict[str, Union[str, int, float]] = Factory(dict)
    data: Dict[str, str] = Factory(dict)
    auth_type: str = ""
    auth_str: List[str] = Factory(list)
    sub_type: str = ""


class HttpProvider(Provider):
    """
    HTTP Generic lookup provider base class.

    For subclasses:

    Define Base URL of the service

    .. code:: python

        _BASE_URL = "https://my.api.org/"

    Define query parameters for different item types (keys)

    ..code:: python

        _QUERIES: Dict[str, APILookupParams] = {}

    For example:

    .. code:: python

        _QUERIES = {
        # Community API
        "ipv4": APILookupParams(
            path="/v3/community/{observable}",
            headers={"key": "{AuthKey}"},
        ),
        # Enterprise API Quick Lookup
        "ipv4-quick": APILookupParams(
            ...


    Define list of required __init__ params

    .. code:: python

        _REQUIRED_PARAMS: List[str] = []

    For example:

    .. code:: python

        _REQUIRED_PARAMS = ["AuthKey"]

    In __init__

    Be sure to call

    .. code::

        super().__init__(**kwargs)``

    Supply any additional checkers/pre-processors
    with

    .. code: python

        self._preprocessors.add_check(type, check_func)

    See Also
    --------
    PreProcessor
    HttpTIProvider

    """

    # Base URL of the service
    _BASE_URL = ""

    # Define query parameters for different item types (keys)
    _QUERIES: Dict[str, APILookupParams] = {}

    # List of required __init__ params
    _REQUIRED_PARAMS: List[str] = []

    def __init__(self, **kwargs):
        """Initialize the class."""
        super().__init__(**kwargs)
        self._httpx_client = httpx.Client(timeout=get_http_timeout(**kwargs))
        self._request_params = {}
        if "ApiID" in kwargs:
            api_id = kwargs.pop("ApiID")
            self._request_params["ApiID"] = api_id.strip() if api_id else None
        if "AuthKey" in kwargs:
            auth_key = kwargs.pop("AuthKey")
            self._request_params["AuthKey"] = auth_key.strip() if auth_key else None
        if "Instance" in kwargs:
            auth_key = kwargs.pop("Instance")
            self._request_params["Instance"] = auth_key.strip() if auth_key else None

        missing_params = [
            param
            for param in self._REQUIRED_PARAMS
            if param not in self._request_params
        ]

        missing_params = []

        if missing_params:
            param_list = ", ".join(f"'{param}'" for param in missing_params)
            raise MsticpyConfigError(
                f"Parameter values missing for Provider '{self.__class__.__name__}'",
                f"Missing parameters are: {param_list}",
            )

        # In __init__ you might want to
        # supply additional checkers/preprocessors
        # with
        # self._preprocessors.add_check(type, check_func)
        # or replace the default PreProcessors
        # self._preprocessors = MyPreProcessor()

    @abstractmethod
    def lookup_item(
        self, item: str, item_type: str = None, query_type: str = None, **kwargs
    ) -> pd.DataFrame:
        """
        Lookup from an item value.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            The Type of the item to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """

    # pylint: enable=duplicate-code
    def _substitute_parms(
        self, value: str, value_type: str, query_type: str = None
    ) -> Tuple[str, Dict[str, Any]]:
        """
        Create requests parameters collection.

        Parameters
        ----------
        value : str
            The value of the item being queried
        value_type : str, optional
            The value type, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the type
            will be returned.

        Returns
        -------
        Tuple[str, Dict[str, Any]]
            HTTP method, dictionary of parameter keys/values

        """
        req_params = {"observable": value}
        req_params.update(self._request_params)
        value_key = f"{value_type}-{query_type}" if query_type else value_type
        src = self.item_query_defs.get(value_key, None)
        if not src:
            raise LookupError(f"Provider does not support this type {value_key}.")

        # create a parameter dictionary to pass to requests
        # substitute any parameter value from our req_params dict
        req_dict: Dict[str, Any] = {
            "headers": {},
            "url": (
                src.path.format(**req_params)
                if src.full_url
                else (self._BASE_URL + src.path).format(**req_params)
            ),
        }

        if src.headers:
            headers: Dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.headers.items()
            }
            req_dict["headers"] = headers
        if "User-Agent" not in req_dict["headers"]:
            req_dict["headers"].update(mp_ua_header())
        if src.params:
            q_params: Dict[str, Any] = {
                key: val.format(**req_params) if isinstance(val, str) else val
                for key, val in src.params.items()
            }
            req_dict["params"] = q_params
        if src.data:
            q_data: Dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.data.items()
            }
            req_dict["data"] = q_data
        if src.auth_type and src.auth_str:
            auth_strs: Tuple = tuple(p.format(**req_params) for p in src.auth_str)
            if src.auth_type == "HTTPBasic":
                req_dict["auth"] = auth_strs
            else:
                raise NotImplementedError(f"Unknown auth type {src.auth_type}")
        return src.verb, req_dict

    @staticmethod
    def _failed_response(response: Dict) -> bool:
        """
        Return True if negative response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        bool
            True if the response indicated failure.

        """
        return (
            response["Status"] not in (200, LookupStatus.OK.value)
            or not response["RawResult"]
            or not isinstance(response["RawResult"], dict)
        )

    @staticmethod
    def _err_to_results(result: Dict, err: Exception):
        result["Details"] = err.args
        result["RawResult"] = (
            type(err).__name__ + "\n" + str(err) + "\n" + traceback.format_exc()
        )

    @staticmethod
    def _response_message(status_code):
        if status_code == 404:
            return "Not found."
        if status_code == 401:
            return "Authorization failed. Check account and key details."
        if status_code == 403:
            return "Request forbidden. Allowed query rate may have been exceeded."
        return httpx.codes.get_reason_phrase(status_code) or "Unknown HTTP status code."



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Utility sub-package."""
from ..._version import VERSION
from .format import *  # noqa: F401, F403
from .package import *  # noqa: F401, F403
from .package import _MSTICPY_USER_AGENT

# All modules use the "export" decorator to control
# pylint: disable=unused-import
from .types import export, singleton  # noqa: F401

from .types import *  # isort: skip  # noqa: F401, F403
from .ipython import *  # isort: skip  # noqa: F401, F403

MSTICPY_USER_AGENT = _MSTICPY_USER_AGENT


__version__ = VERSION
__author__ = "Ian Hellen"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Local Osquery Data Driver class - osquery.{results,snapshots}.log."""
import json
import logging

# pickle used for caching data. However, it is not the default. The
# user would have to specify or configure a cache file to be read.
import pickle  # nosec
import re
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pandas import to_datetime, to_numeric
from tqdm.auto import tqdm

from ..._version import VERSION
from ...common.exceptions import MsticpyDataQueryError, MsticpyNoDataSourceError
from ...common.provider_settings import get_provider_settings
from ...common.utility import export, valid_pyname
from .driver_base import DriverBase, QuerySource

__version__ = VERSION
__author__ = "juju4"

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
@export
class OSQueryLogDriver(DriverBase):
    """OSQueryLogDriver class to execute kql queries."""

    OS_QUERY_DATEIME_COLS = {
        "unixTime",
        "columns_time",
        "columns_atime",
        "columns_ctime",
        "columns_mtime",
    }

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiate OSQueryLogDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string (not used)
        data_paths : List[str], optional
            Paths from which to load data files
        cache_file : str, optional
            Store extracted data to `cache_file` path, or
            read from this file, if it exists.
        progress : bool, optional
            Show progress with tqdm, by default, True

        """
        del connection_str
        self._debug = kwargs.get("debug", False)
        super().__init__()
        self._cache_file: Optional[str] = None
        self._paths: List[str] = ["."]
        # If data paths specified, use these
        # from kwargs or settings
        if data_paths := kwargs.get("data_paths"):
            self._paths = [path.strip() for path in data_paths]
            logger.info("data paths read from param %s", str(self._paths))
        else:
            prov_settings = get_provider_settings(config_section="DataProviders").get(
                "OSQueryLogs"
            )
            if prov_settings:
                self._paths = prov_settings.args.get("data_paths", []) or self._paths
                self._cache_file = prov_settings.args.get("cache_file")
                logger.info("data paths read from config %s", str(self._paths))

        self._progress = kwargs.pop("progress", True)
        self.data_files: Dict[str, str] = self._get_logfile_paths()
        self._schema: Dict[str, Any] = {}
        self._data_cache: Dict[str, pd.DataFrame] = {}
        self._query_map: Dict[str, List[str]]
        self._cache_file = kwargs.pop("cache_file", self._cache_file)
        self._loaded = True
        self.has_driver_queries = True
        logger.info("data files to read %s", ",".join(self.data_files.values()))
        logger.info("cache file %s", self._cache_file)

    def _get_logfile_paths(self) -> Dict[str, str]:
        """Read files in data paths."""
        data_files = {}
        for input_path in (Path(path_str) for path_str in self._paths):
            if input_path.is_file():
                data_files[str(input_path.name).casefold()] = str(input_path)
                continue
            if input_path.is_dir():
                for pattern in ["**/*.log"]:
                    found_files = Path(input_path).resolve().glob(pattern)
                    data_files.update(
                        {
                            str(file_path.name).casefold(): str(file_path)
                            for file_path in found_files
                            if file_path.is_file()
                        }
                    )
        return data_files

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connect to a data source

        """
        del connection_str
        self._read_data_files()
        self._connected = True
        print("Data loaded.")

    # pylint: disable=too-many-branches
    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        if not self._schema:
            self.connect()
            self._schema = {
                table: {col: dtype.name for col, dtype in df.dtypes.to_dict().items()}
                for table, df in self._data_cache.items()
            }

        return self._schema

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        del kwargs
        if not self._data_cache:
            self.connect()
        query_name = query_source.name if query_source else query
        if query_name not in self._query_map:
            raise MsticpyDataQueryError(
                f"No data loaded for query {query_name}.",
                "Please check that the data loaded from the log paths",
                "has this data type.",
            )
        df_names = self._query_map[query_name]
        query_df = pd.concat([self._data_cache[df] for df in df_names])
        for date_column in self.OS_QUERY_DATEIME_COLS & set(query_df.columns):
            query_df[date_column] = to_datetime(
                query_df[date_column], unit="s", origin="unix"
            )
        logger.info("Query %s, returned %d rows", query_name, len(query_df))
        return query_df

    def query_with_results(self, query, **kwargs):
        """Return query with fake results."""
        return self.query(query, **kwargs), "OK"

    @property
    def driver_queries(self) -> List[Dict[str, Any]]:
        """
        Return dynamic queries available on connection to data.

        Returns
        -------
        List[Dict[str, Any]]
            List of queries with properties: "name", "query", "container"
            and (optionally) "description"

        Raises
        ------
        MsticpyNotConnectedError
            If called before driver is connected.

        """
        if not self.connected:
            self.connect()
        if self._query_map:
            return [
                {
                    "name": query,
                    "query": query,
                    "query_paths": "osquery",
                    "description": f"OS Query {query} table.",
                }
                for query in self._query_map
            ]
        return []

    def _read_data_files(self):
        """Read data files into memory."""
        if self._data_cache:
            return
        # If a cache file is specified and is a valid file,
        # try to use that.
        if self._cache_file and Path(self._cache_file).is_file():
            self._read_from_cache()
            logger.info("data read from cache %s", self._cache_file)
            if self._data_cache:
                return

        # Otherwise read in the data files.
        data_files = (
            tqdm(self.data_files.values())
            if self._progress
            else self.data_files.values()
        )
        for log_file in data_files:
            self._read_log_file(log_file)

        # if a cache file name was specified, write data cache to that file.
        if self._cache_file:
            self._write_to_cache()
            logger.info("data written to  %s", self._cache_file)

    def _write_to_cache(self):
        if Path(self._cache_file).is_file():
            print("Overwriting exist cache file.")
        with open(self._cache_file, "wb") as cache_handle:
            try:
                pickle.dump(self._data_cache, cache_handle)
            except pickle.PickleError:
                print(f"Could not write to cache file {self._data_cache}")

    def _read_from_cache(self):
        with open(self._cache_file, "rb") as cache_handle:
            try:
                self._data_cache = pickle.load(cache_handle)  # nosec
            except pickle.PickleError:
                print(f"Invalid cache file {self._data_cache}")

    def _extract_event_type(self, df_all_queries: pd.DataFrame, event_name: str):
        """Extract tidied DF from all queries."""
        query_df = df_all_queries[df_all_queries["name"] == event_name].dropna(
            axis=1, how="all"
        )
        for date_column in self.OS_QUERY_DATEIME_COLS & set(query_df.columns):
            query_df[date_column] = to_datetime(
                to_numeric(query_df[date_column]),
                origin="unix",
                utc=True,
            )
        return query_df

    def _read_log_file(self, log_path: str):
        """Read log file into cache."""
        if Path(log_path).suffix.casefold() not in (".log", ".json"):
            return
        # We can't use directly read_json as we need to extract nested json
        # Split log files in multiple queries table here
        # Likely resource intensive and better way to do.
        # Likely issue, multiple log files can contain same query mostly
        # because of log rotation
        list_lines: List[Dict[str, Any]] = []
        try:
            with open(log_path, mode="r", encoding="utf-8") as logfile:
                json_lines = logfile.readlines()
                list_lines = [json.loads(line) for line in json_lines]

        except (IOError, json.JSONDecodeError, ValueError) as exc:
            raise MsticpyDataQueryError(
                f"Read error on file {log_path}: {exc}."
            ) from exc
        if not list_lines:
            raise MsticpyNoDataSourceError(
                f"No log data retrieved from {log_path}",
                "Please check the format of the log file.",
            )
        logger.info("log %s read, %d lines read", log_path, len(list_lines))
        df_all_queries = pd.json_normalize(list_lines, max_level=3)
        # Don't want dot in columns name
        df_all_queries.columns = df_all_queries.columns.str.replace(
            ".", "_", regex=False
        )

        for event_name in df_all_queries["name"].unique().tolist():
            combined_dfs = []
            # we might already have events of this type in the cache
            if event_name in self._data_cache:
                # if so, we combine with newly-extracted data
                combined_dfs.append(self._data_cache[event_name])
            combined_dfs.append(self._extract_event_type(df_all_queries, event_name))
            # update the cache
            self._data_cache[event_name] = _rename_columns(pd.concat(combined_dfs))
            logger.info(
                "event type %s read to _data_cache, len: %d",
                event_name,
                len(self._data_cache[event_name]),
            )

        self._query_map = self._get_table_names()

    _OSQ_TABLE_REGEX = r"(?:pack_[^_]+_)?(?P<table_name>[\w_]+)"

    def _get_table_names(self):
        """Extract simple table names from the data_cache entries."""
        table_map = defaultdict(list)
        for name in self._data_cache:
            match = re.search(self._OSQ_TABLE_REGEX, name, re.IGNORECASE)
            table_name = match.groupdict().get("table_name") if match else name
            table_name = valid_pyname(table_name.strip())
            table_map[table_name].append(name)
        return table_map


_PREFIXES = ("decorations_", "columns_")


def _rename_columns(data: pd.DataFrame):
    """Rename nested columns."""
    df_cols = set(data.columns)
    rename_cols = {}
    for prefix in _PREFIXES:
        source_cols = data.filter(regex=f"{prefix}.*").columns
        rename_cols.update(
            {
                col: col.replace(prefix, "")
                for col in source_cols
                if isinstance(col, str)
            }
        )
    rename_cols = {
        col: ren_col if ren_col not in df_cols else f"{ren_col}_"
        for col, ren_col in rename_cols.items()
    }
    return data.rename(columns=rename_cols)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Pivot initialization sub-package.

This is a sub-package that handles the pivot initialization.

- pivot_data_queries
- pivot_ti_providers
- vt_pivot

"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Check current version against PyPI."""
import contextlib

import httpx
from pkg_resources import parse_version

from .._version import VERSION
from .utility import mp_ua_header, unit_testing

__version__ = VERSION
__author__ = "Ian Hellen"


def check_version():
    """Check the current version against latest on PyPI."""
    installed_version = parse_version(__version__)

    # fetch package metadata from PyPI
    pypi_url = "https://pypi.org/pypi/msticpy/json"
    pkg_data = {"info": {"version": "0.0.0"}, "releases": {}}
    with contextlib.suppress(httpx.ConnectError):
        if not unit_testing():
            resp = httpx.get(
                pypi_url,
                timeout=httpx.Timeout(2.0, connect=2.0),
                headers=mp_ua_header(),
            )
            if resp.status_code == 200:
                pkg_data = resp.json()

    latest_version = pkg_data.get("info", {}).get("version", None)
    if latest_version:
        latest_version = parse_version(latest_version)
    elif "releases" in pkg_data:
        latest_version = max(parse_version(s) for s in pkg_data["releases"].keys())

    print(
        "msticpy version",
        "installed:",
        installed_version,
        "latest published:",
        latest_version if str(latest_version) != "0.0.0" else "unknown",
    )
    if installed_version < latest_version:
        print(f"A newer version of msticpy - {latest_version} is available.")
        print("Upgrade with 'pip install --upgrade msticpy'")
    else:
        print("Latest known version is installed.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP TI Provider base.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from functools import lru_cache
from json import JSONDecodeError
from typing import Any, Dict

import pandas as pd

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ...common.utility import export
from ..http_provider import HttpProvider
from ..lookup_result import LookupStatus
from .result_severity import ResultSeverity
from .ti_provider_base import TIProvider

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class HttpTIProvider(TIProvider, HttpProvider):
    """HTTP API Lookup provider base class."""

    @lru_cache(maxsize=256)
    def lookup_ioc(
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        ioc : str
            ioc to lookup
        ioc_type : str, optional
            The Type of the ioc to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the ioc
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """
        result = self._check_ioc_type(ioc, ioc_type, query_subtype=query_type)

        result["Provider"] = kwargs.get("provider_name", self.__class__.__name__)

        req_params: Dict[str, Any] = {}
        try:
            verb, req_params = self._substitute_parms(
                result["SafeIoc"], result["IocType"], query_type
            )
            if verb == "GET":
                response = self._httpx_client.get(
                    **req_params, timeout=get_http_timeout(**kwargs)
                )
            elif verb == "POST":
                response = self._httpx_client.post(
                    **req_params, timeout=get_http_timeout(**kwargs)
                )
            else:
                raise NotImplementedError(f"Unsupported verb {verb}")

            result["Status"] = response.status_code
            result["Reference"] = req_params["url"]
            if result["Status"] == 200:
                try:
                    result["RawResult"] = response.json()
                    result["Result"], severity, result["Details"] = self.parse_results(
                        result
                    )
                except JSONDecodeError:
                    result[
                        "RawResult"
                    ] = f"""There was a problem parsing results from this lookup:
                                        {response.text}"""
                    result["Result"] = False
                    severity = ResultSeverity.information
                    result["Details"] = {}
                if isinstance(severity, ResultSeverity):
                    result["Severity"] = severity.name
                result["Severity"] = ResultSeverity.parse(severity).name
                result["Status"] = LookupStatus.OK.value
            else:
                result["RawResult"] = str(response)
                result["Result"] = False
                result["Details"] = self._response_message(result["Status"])
        except (
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            self._err_to_results(result, err)
            if not isinstance(err, LookupError):
                url = req_params.get("url", None) if req_params else None
                result["Reference"] = url
        return pd.DataFrame([result])



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for TILookup classes.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import contextlib
import math  # noqa
import re
from collections import Counter
from functools import partial
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Callable, List, Optional, Set, Tuple, Union
from urllib.parse import quote_plus

from urllib3.exceptions import LocationParseError
from urllib3.util import parse_url

from .._version import VERSION
from ..common.utility.format import refang_ioc
from ..transform.iocextract import IoCExtract
from .lookup_result import SanitizedObservable

__version__ = VERSION
__author__ = "Ian Hellen"

_IOC_EXTRACT = IoCExtract()


# slightly stricter than normal URL regex to exclude '() from host string
_HTTP_STRICT_REGEX = r"""
    (?P<protocol>(https?|ftp|telnet|ldap|file)://)
    (?P<userinfo>([a-z0-9-._~!$&*+,;=:]|%[0-9A-F]{2})*@)?
    (?P<host>([a-z0-9-._~!$&\*+,;=]|%[0-9A-F]{2})*)
    (:(?P<port>\d*))?
    (/(?P<path>([^?\#| ]|%[0-9A-F]{2})*))?
    (\?(?P<query>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?
    (\#(?P<fragment>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?\b"""

_HTTP_STRICT_RGXC = re.compile(_HTTP_STRICT_REGEX, re.I | re.X | re.M)


CheckerType = Callable[..., SanitizedObservable]


def _preprocess_url(url: str, **kwargs) -> SanitizedObservable:
    """
    Check that URL can be parsed.

    Parameters
    ----------
    url : str
        The URL to check
    require_url_encoding : bool
        Set to True if url's require encoding before passing to provider

    Returns
    -------
    SanitizedObservable
        Pre-processed result

    """
    require_url_encoding: bool = kwargs.pop("require_url_encoding", False)
    url = refang_ioc(ioc=url, ioc_type="url")
    clean_url, scheme, host = get_schema_and_host(url, require_url_encoding)

    if scheme is None or host is None:
        return SanitizedObservable(None, f"Could not obtain scheme or host from {url}")
    # get rid of some obvious false positives (localhost, local hostnames)
    with contextlib.suppress(ValueError):
        addr = ip_address(host)
        if addr.is_private:
            return SanitizedObservable(None, "Host part of URL is a private IP address")
        if addr.is_loopback:
            return SanitizedObservable(
                None, "Host part of URL is a loopback IP address"
            )
    if "." not in host:
        return SanitizedObservable(None, "Host is unqualified domain name")

    if scheme.lower() in ["file"]:
        return SanitizedObservable(None, f"{scheme} URL scheme is not supported")

    return SanitizedObservable(clean_url, "ok")


def get_schema_and_host(
    url: str, require_url_encoding: bool = False
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Return URL scheme and host and cleaned URL.

    Parameters
    ----------
    url : str
        Input URL
    require_url_encoding : bool
        Set to True if url needs encoding. Default is False.

    Returns
    -------
    Tuple[Optional[str], Optional[str], Optional[str]
        Tuple of URL, scheme, host

    """
    clean_url = None
    scheme = None
    host = None
    try:
        scheme, _, host, _, _, _, _ = parse_url(url)
        clean_url = url
    except LocationParseError:
        # Try to clean URL and re-check
        cleaned_url = _clean_url(url)
        if cleaned_url is not None:
            with contextlib.suppress(LocationParseError):
                scheme, _, host, _, _, _, _ = parse_url(cleaned_url)
                clean_url = cleaned_url
    if require_url_encoding and clean_url:
        clean_url = quote_plus(clean_url)
    return clean_url, scheme, host


def _clean_url(url: str) -> Optional[str]:
    """
    Clean URL to remove query params and fragments and any trailing stuff.

    Parameters
    ----------
    url : str
        the URL to check

    Returns
    -------
    Optional[str]
        Cleaned URL or None if the input was not a valid URL

    """
    # Try to clean URL and re-check
    match_url = _HTTP_STRICT_RGXC.search(url)
    if (
        not match_url
        or match_url.groupdict()["protocol"] is None
        or match_url.groupdict()["host"] is None
    ):
        return None

    # build the URL dropping the query string and fragments
    clean_url = match_url.groupdict()["protocol"]
    if match_url.groupdict()["userinfo"]:
        clean_url += match_url.groupdict()["userinfo"]
    clean_url += match_url.groupdict()["host"]
    if match_url.groupdict()["port"]:
        clean_url += ":" + match_url.groupdict()["port"]
    if match_url.groupdict()["path"]:
        clean_url += "/" + match_url.groupdict()["path"]

    return clean_url


def _preprocess_ip(ipaddress: str, **kwargs):
    """Ensure Ip address is a valid public IPv4 address."""
    version = kwargs.pop("version", 4)
    ipaddress = refang_ioc(ioc=ipaddress, ioc_type="ipv4")
    try:
        addr = ip_address(ipaddress)
    except ValueError:
        return SanitizedObservable(None, "IP address is invalid format")

    if version == 4 and not isinstance(addr, IPv4Address):
        return SanitizedObservable(None, "Not an IPv4 address")
    if version == 6 and not isinstance(addr, IPv6Address):
        return SanitizedObservable(None, "Not an IPv6 address")
    if addr.is_global:
        return SanitizedObservable(ipaddress, "ok")

    return SanitizedObservable(None, "IP address is not global")


def _preprocess_dns(domain: str, **kwargs) -> SanitizedObservable:
    """Ensure DNS is a valid-looking domain."""
    del kwargs
    domain = refang_ioc(ioc=domain, ioc_type="dns")
    if "." not in domain:
        return SanitizedObservable(None, "Domain is unqualified domain name")
    with contextlib.suppress(ValueError):
        addr = ip_address(domain)
        del addr
        return SanitizedObservable(None, "Domain is an IP address")
    return SanitizedObservable(domain, "ok")


def _preprocess_hash(hash_str: str, **kwargs) -> SanitizedObservable:
    """Ensure Hash has minimum entropy (rather than a string of 'x')."""
    del kwargs
    str_entropy = _entropy(hash_str)
    if str_entropy < 3.0:
        return SanitizedObservable(None, "String has too low an entropy to be a hash")
    return SanitizedObservable(hash_str, "ok")


def _validate_ioc_type(observable, ioc_type):
    """Validate type matches IoCExtract regex rules."""
    try:
        validated = _IOC_EXTRACT.validate(observable, ioc_type)
    except KeyError:
        validated = False
    if not validated:
        return SanitizedObservable(
            None, f"Observable does not match expected pattern for {ioc_type}"
        )
    return SanitizedObservable(observable, "ok")


class PreProcessor:
    """Observable pre-processing class."""

    _TYPE_CHECK = "type_check"

    # Default processors
    _DEF_PROCESSORS: List[Tuple[Set[str], List[Union[str, CheckerType]]]] = [
        ({"url"}, [_TYPE_CHECK, _preprocess_url]),
        ({"ipv4"}, [_TYPE_CHECK, _preprocess_ip]),
        ({"ipv6"}, [_TYPE_CHECK, partial(_preprocess_ip, version=6)]),
        ({"dns", "hostname"}, [_TYPE_CHECK, _preprocess_dns]),
        (
            {"md5_hash", "sha1_hash", "sha256_hash", "file_hash"},
            [_TYPE_CHECK, _preprocess_hash],
        ),
    ]

    def __init__(self):
        """Initialize the processor dictionary."""
        self._processors = {
            obs_type: processors
            for types, processors in self._DEF_PROCESSORS
            for obs_type in types
        }

    def check(self, value: str, value_type: str, **kwargs) -> SanitizedObservable:
        """
        Apply processing checks to the input value.

        Parameters
        ----------
        value : str
            The value to be checked.
        value_type : str
            The type of value to be checked.

        Returns
        -------
        SanitizedObservable
            Named tuple with two members:

            1. observable - with the pre-processed result,
               This is set to None if a check fails.
            2. status - this is set to "ok" if the checks completed.
               Otherwise, it has an error message.

        """
        proc_value: Optional[str] = value.strip()
        result = SanitizedObservable(proc_value, "ok")
        for processor in self._processors.get(value_type, []):
            if processor == self._TYPE_CHECK:
                result = _validate_ioc_type(proc_value, value_type)
            else:
                result = processor(proc_value, **kwargs)
            proc_value = result.observable
            if proc_value is None:
                break
        return result

    def add_check(self, value_type: str, checker: CheckerType):
        """Add a new checker to the processors."""
        if value_type not in self._processors:
            self._processors[value_type] = [checker]
        else:
            self._processors[value_type].append(checker)


def preprocess_observable(
    observable, ioc_type, require_url_encoding: bool = False
) -> SanitizedObservable:
    """
    Preprocess and check validity of observable against declared IoC type.

    Parameters
    ----------
    observable : _type_
        the value of the observable
    ioc_type : _type_
        The type of observable
    require_url_encoding : bool, optional
        If the observable needs URL-encoding (URL types only), by default False

    Returns
    -------
    SanitizedObservable
        Named tuple with two members:

        1. observable - with the pre-processed result,
            This is set to None if a check fails.
        2. status - this is set to "ok" if the checks completed.
            Otherwise, it has an error message.

    """
    processor = PreProcessor()
    return processor.check(
        value=observable, value_type=ioc_type, require_url_encoding=require_url_encoding
    )


def _entropy(input_str: str) -> float:
    """Compute entropy of input string."""
    str_len = float(len(input_str))
    return -sum(
        map(
            lambda a: (a / str_len) * math.log2(a / str_len),
            Counter(input_str).values(),
        )
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Msticpy Config class."""

import io
import pprint
from contextlib import redirect_stdout, suppress
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Union

import ipywidgets as widgets
import yaml
from IPython.display import display

from .._version import VERSION

try:
    from ..auth.keyvault_client import BHKeyVaultClient, MsticpyKeyVaultConfigError

    _KEYVAULT = True
except ImportError:
    _KEYVAULT = False

from ..common.pkg_config import (
    SettingsDict,
    current_config_path,
    refresh_config,
    validate_config,
)
from ..common.utility.package import delayed_import
from .comp_edit import CompEditDisplayMixin, CompEditStatusMixin
from .file_browser import FileBrowser

ms_sentinel = delayed_import("msticpy.context.azure.sentinel_core", "MicrosoftSentinel")

__version__ = VERSION
__author__ = "Ian Hellen"


_CONFIG_MAP = {
    "resource_group": "ResourceGroup",
    "subscription_id": "SubscriptionId",
    "tenant_id": "TenantId",
    "workspace_id": "WorkspaceId",
    "workspace_name": "WorkspaceName",
}

_TXT_STYLE = {
    "style": {"description_width": "150px"},
    "layout": widgets.Layout(width="99%"),
}


# pylint: disable=too-many-instance-attributes
class MpConfigFile(CompEditStatusMixin, CompEditDisplayMixin):
    """
    MSTICPy Configuration management class.

    Use the functions from the commandline or display
    in a Jupyter notebook to use interactive version.

    """

    _DEF_FILENAME = "./msticpyconfig.yaml"

    def __init__(
        self,
        file: Union[str, Path, None] = None,
        settings: Union[Dict[str, Any], SettingsDict, None] = None,
    ):
        """
        Create an instance of the MSTICPy Configuration helper class.

        Parameters
        ----------
        file : Optional[str], optional
            config file path to use.
            If `file` is not supplied, the msticpyconfig path defaults to the system
            default location for msticpyconfig.yaml (MSTICPYCONFIG env var,
            current directory, home directory). This path will also be used as
            the location to save any changed settings.
            If `file` is supplied and exists, the configuration will be loaded from here,
            unless `settings` parameter is also is supplied. If `settings` is supplied
            the `file` path is used as the file to save configuration but
            the configuration settings are taken from the `settings` parameter
            If `file` is supplied but does not exist, no configuration
            is loaded but this path will be used as the file name to
            save any configuration by default None
        settings : Optional[Dict[str, Any]], optional
            setting dict to load, by default None
            If a settings dictionary is supplied it will override any settings
            read from the `file` parameter.

        """
        self.kv_client: Any = None

        # Set up controls
        self.file_browser = FileBrowser(select_cb=self.load_from_file)
        self.txt_viewer = widgets.Textarea(
            layout=widgets.Layout(width="99%", height="300px")
        )
        self.btn_close = widgets.Button(description="Close viewer")
        self.btn_close.on_click(self._close_view)

        self.html_title = widgets.HTML("<h3>MSTICPy settings</h3>")
        self.txt_current_config_path = widgets.Text(
            description="Current file", **_TXT_STYLE
        )
        self.txt_default_config_path = widgets.Text(
            description="Default Config path", **_TXT_STYLE
        )
        self._txt_import_url = widgets.Text(
            description="MS Sentinel Portal URL", **_TXT_STYLE
        )
        self._last_workspace: Dict[str, Dict[str, str]]
        self.buttons: Dict[str, widgets.Button] = {}
        self.btn_pane = self._setup_buttons()
        self.info_pane = widgets.VBox(
            [
                self.txt_current_config_path,
                self.txt_default_config_path,
                self._txt_import_url,
            ],
            layout=self.border_layout("80%"),
        )
        self.viewer = widgets.VBox([])
        self.edit_controls = widgets.VBox([self.btn_pane, self.info_pane])
        self.layout = widgets.VBox(
            [
                self.html_title,
                self.edit_controls,
                self.viewer,
            ],
            layout=self.border_layout("99%"),
        )

        if file is None:
            self.current_file = current_config_path()
            if self.current_file is None:
                self.current_file = self._DEF_FILENAME
        else:
            self.current_file = file

        # set the default location even if user supplied file parameter
        self.mp_config_def_path = current_config_path() or self.current_file

        if settings is not None and isinstance(settings, (dict, SettingsDict)):
            # If caller has supplied settings, we don't want to load
            # anything from a file
            self.settings = SettingsDict(settings)
            return
        # otherwise load settings from the current config file.
        if self.current_file and Path(self.current_file).is_file():
            self.load_from_file(self.current_file)
        else:
            # no file so set default empty settings
            self.settings = SettingsDict()
            self.set_status(f"Filename does not exist: '{self.current_file}'.")

    @property
    def current_file(self):
        """Return currently loaded file path."""
        return self.txt_current_config_path.value

    @current_file.setter
    def current_file(self, file_name: Union[str, Path]):
        """Set currently loaded file path."""
        self.txt_current_config_path.value = str(file_name)

    @property
    def default_config_file(self):
        """Return default msticpyconfig path."""
        return self.txt_default_config_path.value

    @default_config_file.setter
    def default_config_file(self, file_name: Union[str, Path]):
        """Set default msticpyconfig path."""
        self.txt_default_config_path.value = file_name

    def load_default(self):
        """Load default settings specified by MSTICPYCONFIG env var."""
        if self.mp_config_def_path:
            self.current_file = self.mp_config_def_path
            with suppress(FileNotFoundError):
                self.load_from_file(self.mp_config_def_path)

    def browse_for_file(self, show: bool = True):
        """Open the browser to browser/search for a file."""
        self.viewer.children = [self.file_browser.layout, self.btn_close]
        if show:
            display(self.viewer)

    def load_from_file(self, file: Union[str, Path]):
        """Load settings from `file`."""
        self.settings = self._read_mp_config(file)
        self.current_file = file

    def view_settings(self, show: bool = True):
        """View the current settings as text."""
        self.txt_viewer.value = pprint.pformat(self.settings, compact=True)
        self.viewer.children = [self.txt_viewer, self.btn_close]
        if show:
            display(self.viewer)

    def _close_view(self, btn):
        del btn
        self.viewer.children = []

    def validate_settings(self, show: bool = True):
        """Run the validator against currently loaded settings."""
        results = io.StringIO()
        with redirect_stdout(results):
            validate_config(mp_config=self.settings)
        self.txt_viewer.value = results.getvalue()
        self.viewer.children = [self.txt_viewer, self.btn_close]
        if show:
            display(self.viewer)

    def save_to_file(self, file: str, backup: bool = True):
        """
        Save current configuration to `file`.

        Parameters
        ----------
        file : str
            The file path to save to.
        backup : bool, optional
            Create a backup file, if overwriting existing file,
            by default True

        """
        # remove empty settings sections before saving
        empty_items = [
            section for section, settings in self.settings.items() if not settings
        ]
        for empty_section in empty_items:
            del self.settings[empty_section]
        # create a backup, if required
        if backup and Path(file).is_file():
            Path(file).replace(f"{file}.save_{datetime.now().strftime('%H%M%S')}")
        with open(file, "w", encoding="utf-8") as mp_hdl:
            yaml.safe_dump(dict(self.settings), mp_hdl)

    def show_kv_secrets(self, show: bool = True):
        """Show secrets from currently configured Key Vault."""
        view_text = []
        if not _KEYVAULT:
            self.txt_viewer.value = "\n".join(
                [
                    "Azure keyvault libraries not found.",
                    "Please install 'azure_keyvault_secrets' and",
                    "'azure_mgmt_keyvault'",
                ]
            )
            self.viewer.children = [self.txt_viewer, self.btn_close]
            if show:
                display(self.viewer)
            return
        if self.kv_client is None:
            try:
                self.kv_client = BHKeyVaultClient()
            except MsticpyKeyVaultConfigError:
                view_text = ["Key Vault settings have not been configured correctly."]
        if self.kv_client is not None:
            try:
                secrets = self.kv_client.secrets
            except Exception as err:  # pylint: disable=broad-except
                view_text.extend(
                    [
                        "Could not get secrets list from Key Vault.",
                        "It is likely that the authentication failed."
                        f"The exception is show here: {err}",
                    ]
                )
                secrets = []

            for sec_path in secrets:
                sec_name = sec_path.rsplit("/", maxsplit=1)[-1]
                try:
                    sec_result = f"Value: '{self.kv_client.get_secret(sec_name)}'"
                except Exception:  # pylint: disable=broad-except
                    sec_result = "Value: Could not display secret"
                view_text.append(f"Secret '{sec_name}'  {sec_result}")
        self.txt_viewer.value = "\n".join(view_text)
        self.viewer.children = [self.txt_viewer, self.btn_close]
        if show:
            display(self.viewer)

    @staticmethod
    def get_workspace_from_url(url: str) -> Dict[str, Dict[str, str]]:
        """
        Return workspace settings from Sentinel portal URL.

        Parameters
        ----------
        url : str
            The URL of the Azure portal page for the Sentinel workspace.

        Returns
        -------
        Dict[str, Dict[str, str]]
            Dictionary with a single element keyed by workspace name
            The value is the workspace settings dictionary for the
            workspace.

        """
        return ms_sentinel().get_workspace_details_from_url(url)

    def _show_sentinel_workspace(self, show: bool = True):
        """Fetch settings from Sentinel Portal URL."""
        if not self._txt_import_url.value:
            return
        self._last_workspace = self.get_workspace_from_url(self._txt_import_url.value)
        workspace_settings = pprint.pformat(self._last_workspace, compact=True)
        self.txt_viewer.value = "\n".join(
            [
                workspace_settings,
                "\n"
                "Use 'Import into settings' button to import into current settings.",
            ]
        )
        self.viewer.children = [self.txt_viewer, self.btn_close]
        if self._last_workspace:
            self.buttons["import_workspace"].disabled = False
        if show:
            display(self.viewer)

    def _import_sentinel_settings(self):
        """Import sentinel settings from portal URL."""
        if not self._last_workspace:
            return
        curr_workspaces = self.settings.get("MSSentinel", {}).get("Workspaces")
        curr_workspaces.update(self._last_workspace)
        self.view_settings()

    def _read_mp_config(self, file):
        if Path(file).is_file():
            with open(file, "r", encoding="utf-8") as mp_hdl:
                try:
                    return SettingsDict(yaml.safe_load(mp_hdl))
                except yaml.scanner.ScannerError as err:
                    self.set_status(str(err))
        raise FileNotFoundError(f"Cannot read file {file}")

    def map_json_to_mp_ws(self):
        """Map config.json settings to MSTICPy settings."""
        if "resource_group" in self.settings:
            ws_settings = {
                config: self.settings[entry] for entry, config in _CONFIG_MAP.items()
            }
            workspace = self.settings.get("workspace_name", "Default")
            self.settings = SettingsDict(
                {"MSSentinel": {"Workspaces": {workspace: ws_settings}}}
            )
            return self.settings
        return {}

    @staticmethod
    def refresh_mp_config():
        """Refresh global MSTICPy settings from config file."""
        refresh_config()

    def _convert_to_mp(self, btn):
        del btn
        self.map_json_to_mp_ws()
        self.view_settings()

    def _save_file(self, btn):
        del btn
        if self.current_file:
            self.save_to_file(file=self.current_file, backup=True)

    def _btn_func(self, func_name: str):
        """Wrap methods to be called from button events."""

        def _btn_exec(*args):
            del args
            func = getattr(self, func_name)
            func()

        return _btn_exec

    def _btn_func_no_disp(self, func_name: str):
        """Wrap methods to be called from button events."""

        def _btn_exec(*args):
            del args
            func = getattr(self, func_name)
            func(show=False)

        return _btn_exec

    _BUTTON_DEFS = {
        "load": {
            "description": "Load File",
            "tooltip": "Load a settings file from a path.",
        },
        "load_def": {
            "description": "Load default",
            "tooltip": "Load default settings from configured msticpyconfig.yaml.",
        },
        "reload": {
            "description": "Reload",
            "tooltip": "Reload current saved settings to MSTICPy.",
        },
        "view": {
            "description": "View Settings",
            "tooltip": "Display settings in viewer.",
        },
        "convert": {
            "description": "Convert to MP",
            "tooltip": "Convert config.json settings to MSTICPy format.",
        },
        "save": {
            "description": "Save File",
            "tooltip": "Save current settings to current file.",
        },
        "validate": {
            "description": "Validate Settings",
            "tooltip": "Run validation checks on current settings.",
        },
        "showkv": {
            "description": "Show Key Vault secrets",
            "tooltip": "Show Key Vault secrets from current config.",
        },
        "get_workspace": {
            "description": "Get Workspace from URL",
            "tooltip": "Import Sentinel workspace settings from portal URL.",
        },
        "import_workspace": {
            "description": "Import into settings",
            "tooltip": "Import retrieved Workspace settings into current settings.",
            "disabled": True,
        },
    }

    def _setup_buttons(self):
        btn_style = {"layout": widgets.Layout(width="200px")}
        self.buttons["load"] = widgets.Button(
            **(self._BUTTON_DEFS["load"]), **btn_style
        )
        self.buttons["load_def"] = widgets.Button(
            **(self._BUTTON_DEFS["load_def"]), **btn_style
        )
        self.buttons["reload"] = widgets.Button(
            **(self._BUTTON_DEFS["reload"]), **btn_style
        )
        self.buttons["view"] = widgets.Button(
            **(self._BUTTON_DEFS["view"]), **btn_style
        )
        self.buttons["validate"] = widgets.Button(
            **(self._BUTTON_DEFS["validate"]), **btn_style
        )
        self.buttons["convert"] = widgets.Button(
            **(self._BUTTON_DEFS["convert"]), **btn_style
        )
        self.buttons["save"] = widgets.Button(
            **(self._BUTTON_DEFS["save"]), **btn_style
        )
        self.buttons["showkv"] = widgets.Button(
            **(self._BUTTON_DEFS["showkv"]), **btn_style
        )
        self.buttons["get_workspace"] = widgets.Button(
            **(self._BUTTON_DEFS["get_workspace"]), **btn_style
        )
        self.buttons["import_workspace"] = widgets.Button(
            **(self._BUTTON_DEFS["import_workspace"]), **btn_style
        )
        self._btn_view_setting = widgets.Button(
            description="Get Workspace", **btn_style
        )
        self._btn_import_settings = widgets.Button(
            description="Import into settings", disabled=True, **btn_style
        )

        self.buttons["load"].on_click(self._btn_func_no_disp("browse_for_file"))
        self.buttons["load_def"].on_click(self._btn_func("load_default"))
        self.buttons["view"].on_click(self._btn_func_no_disp("view_settings"))
        self.buttons["validate"].on_click(self._btn_func_no_disp("validate_settings"))
        self.buttons["convert"].on_click(self._convert_to_mp)
        self.buttons["save"].on_click(self._save_file)
        self.buttons["reload"].on_click(self._btn_func("refresh_mp_config"))
        self.buttons["showkv"].on_click(self._btn_func_no_disp("show_kv_secrets"))
        self.buttons["get_workspace"].on_click(
            self._btn_func("_show_sentinel_workspace")
        )
        self.buttons["import_workspace"].on_click(
            self._btn_func("_import_sentinel_settings")
        )

        btns1 = widgets.VBox(list(self.buttons.values())[: len(self.buttons) // 2])
        # flake8: noqa: E203
        # conflicts with Black formatting
        btns2 = widgets.VBox(list(self.buttons.values())[len(self.buttons) // 2 :])
        btns_all = widgets.HBox([btns1, btns2])
        return widgets.VBox(
            [widgets.Label(value="Operations"), btns_all],
            # layout=self.border_layout("50%"),
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
from typing import Optional

import ipywidgets as widgets
from ipywidgets import Layout

from .._version import VERSION
from .core import IPyDisplayMixin, RegisteredWidget

__version__ = VERSION
__author__ = "Ian Hellen"


class GetText(RegisteredWidget, IPyDisplayMixin):
    """
    GetText.

    Widget to set text value.
    """

    _NB_PARAMS = {"value": "_value"}

    def __init__(
        self,
        default: Optional[str] = None,
        description: str = "Enter the value: ",
        auto_display: bool = False,
        **kwargs,
    ):
        """
        Create a new instance of GetEnvironmentKey.

        Parameters
        ----------
        default : str
            Default value.
        description : str, optional
            Prompt to display with the text box.
            (the default is "Enter the value: ")
            "prompt" is a alias for this parameter
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        See Also
        --------
        RegisteredWidget

        """
        self._value = default
        description = kwargs.pop("prompt", description)

        self._w_text = widgets.Text(
            description=description,
            layout=Layout(width=kwargs.get("width", "50%")),
            style={"description_width": kwargs.get("description_width", "initial")},
            value=default or kwargs.get("value", ""),
        )

        # Call superclass to register
        super().__init__(id_vals=[default, description], val_attrs=["value"], **kwargs)

        if auto_display:
            self.display()

    @property
    def layout(self):
        """Return underlying widget collection."""
        return self._w_text

    @property
    def value(self):
        """Get the current value of the widget."""
        return self._w_text.value.strip()

    @value.setter
    def value(self, value: str):  # sourcery skip: remove-unnecessary-cast
        """Set the value of the widget."""
        if not isinstance(value, str):
            value = str(value)
        self._w_text.value = value



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Deprecated - module data.azure has moved.

See :py:mod:`msticpy.context.azure`
"""
import warnings

from ..._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"

# pylint: disable=unused-import
from ...context.azure.azure_data import AzureData  # noqa: F401
from ...context.azure.sentinel_core import MicrosoftSentinel  # noqa: F401


WARN_MSSG = (
    "This module has moved to msticpy.context.azure\n"
    "Please change your import to reflect this new location."
    "This will be removed in MSTICPy v2.2.0"
)
warnings.warn(WARN_MSSG, category=DeprecationWarning)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""VirusTotal Pivot functions."""

from enum import Flag, auto
from functools import partial
from typing import Dict, Optional, Tuple, Union

from ..._version import VERSION
from ...common.provider_settings import get_provider_settings
from ...common.utility import enum_parse
from ...context import VT3_AVAILABLE
from ..pivot import Pivot, PivotRegistration

if VT3_AVAILABLE:
    from ...context.vtlookupv3.vtlookupv3 import VTLookupV3

__version__ = VERSION
__author__ = "Ian Hellen"

_USE_PRIVATE_API_KEY = "UseVT3PrivateAPI"


class VTAPIScope(Flag):
    """VT API Type enumeration."""

    PUBLIC = auto()
    PRIVATE = auto()
    ALL = PUBLIC | PRIVATE


FILE_RELATIONSHIPS = {
    "bundled_files": VTAPIScope.PUBLIC,
    "contacted_domains": VTAPIScope.PUBLIC,
    "contacted_ips": VTAPIScope.PUBLIC,
    "contacted_urls": VTAPIScope.PUBLIC,
    "dropped_files": VTAPIScope.PUBLIC,
    "execution_parents": VTAPIScope.PUBLIC,
    "email_attachments": VTAPIScope.PRIVATE,
    "email_parents": VTAPIScope.PRIVATE,
    "embedded_domains": VTAPIScope.PRIVATE,
    "embedded_ips": VTAPIScope.PRIVATE,
    "embedded_urls": VTAPIScope.PRIVATE,
    "sigma_analysis": VTAPIScope.PRIVATE,
    "similar_files": VTAPIScope.PRIVATE,
}

DOMAIN_RELATIONSHIPS = {
    "communicating_files": VTAPIScope.PUBLIC,
    "historical_whois": VTAPIScope.PUBLIC,
    "historical_ssl_certificates": VTAPIScope.PUBLIC,
    "parent": VTAPIScope.PUBLIC,
    "resolutions": VTAPIScope.PUBLIC,
    "subdomains": VTAPIScope.PUBLIC,
    "downloaded_files": VTAPIScope.PRIVATE,
    "ns_records": VTAPIScope.PRIVATE,
    "urls": VTAPIScope.PRIVATE,
}

IP_RELATIONSHIPS = {
    "communicating_files": VTAPIScope.PUBLIC,
    "historical_whois": VTAPIScope.PUBLIC,
    "historical_ssl_certificates": VTAPIScope.PUBLIC,
    "referrer_files": VTAPIScope.PUBLIC,
    "resolutions": VTAPIScope.PUBLIC,
    "subdomains": VTAPIScope.PUBLIC,
    "downloaded_files": VTAPIScope.PRIVATE,
    "urls": VTAPIScope.PRIVATE,
}

URL_RELATIONSHIPS = {
    "last_serving_ip_address": VTAPIScope.PUBLIC,
    "network_location": VTAPIScope.PUBLIC,
    "communicating_files": VTAPIScope.PUBLIC,
    "contacted_domains": VTAPIScope.PUBLIC,
    "contacted_ips": VTAPIScope.PUBLIC,
    "downloaded_files": VTAPIScope.PRIVATE,
    "redirecting_urls": VTAPIScope.PRIVATE,
    "redirects_to": VTAPIScope.PRIVATE,
    "referrer_files": VTAPIScope.PRIVATE,
    "referrer_urls": VTAPIScope.PRIVATE,
}

PIVOT_ENTITY_CATS: Dict[str, Tuple[str, Dict[str, VTAPIScope]]] = {
    "File": ("file", FILE_RELATIONSHIPS),
    "IpAddress": ("ip_address", IP_RELATIONSHIPS),
    "Dns": ("domain", DOMAIN_RELATIONSHIPS),
    "Url": ("url", URL_RELATIONSHIPS),
}

_ENTITY_PROPS = {
    "File": "Sha256",
    "IpAddress": "Address",
    "Dns": "DomainName",
    "Url": "Url",
}

_FUNC_DOC = """
{description}

Parameters
----------
observable: str
    The observable value
limit: int
    Relations limit
all_props : bool, optional
    If True, return all properties, by default False
full_objects : bool, optional
    If True, return the full object rather than just ID links.

Returns
-------
    Relationship Pandas DataFrame with the relationships of the entity

"""


def init():
    """Load VT3 Pivots if vt library is available."""
    if VT3_AVAILABLE:
        add_pivot_functions()


# pylint: disable=no-member
def add_pivot_functions(api_scope: Optional[str] = None):
    """
    Add VT3 relationship functions as pivot functions.

    Parameters
    ----------
    api_scope : Optional[str], optional
        "public" or "private", by default "public"

    """
    ent_funcs = _create_pivots(api_scope)
    for entity, funcs in ent_funcs.items():
        for func_name, func in funcs.items():
            pivot_reg = PivotRegistration(
                func_new_name=func_name,
                input_type="value",
                entity_map={entity: _ENTITY_PROPS[entity]},
                func_input_value_arg="entity_id",
                can_iterate=True,
            )
            Pivot.add_pivot_function(func, pivot_reg=pivot_reg, container="VT")


# pylint: disable=no-member
def _create_pivots(api_scope: Union[str, VTAPIScope, None]):
    if api_scope is None:
        scope = _get_vt_api_scope()
    elif isinstance(api_scope, str):
        scope = enum_parse(VTAPIScope, api_scope) or VTAPIScope.PUBLIC
    else:
        scope = api_scope
    try:
        vt_client = VTLookupV3()
    except (ValueError, AttributeError):
        # Can't initialize VTLookup - don't add the pivot funcs
        return {}

    if not isinstance(scope, VTAPIScope):
        # pylint: disable=not-an-iterable
        scope_names = [f"{name.lower()}" for name in VTAPIScope.__members__]
        raise ValueError(
            f"Unknown API scope {api_scope}",
            f"Valid values are {', '.join(scope_names)}",
        )

    if scope == VTAPIScope.PRIVATE:
        scope = VTAPIScope.ALL
    ent_funcs = {}
    for entity, (vt_type, category) in PIVOT_ENTITY_CATS.items():
        ent_relations = (
            rel for rel, rel_scope in category.items() if rel_scope & scope
        )
        func_dict = {}
        for relationship in ent_relations:
            f_part = partial(
                _get_relationships,
                vt_client=vt_client,
                vt_type=vt_type,
                relationship=relationship,
            )
            f_part.__doc__ = _create_func_doc(entity, relationship)
            func_dict[relationship] = f_part
        ent_funcs[entity] = func_dict
    return ent_funcs


# pylint: enable=no-member


def _create_func_doc(entity, relationship):
    """Create the relationship docstring."""
    fmt_name = " ".join(rel.capitalize() for rel in relationship.split("_"))
    description = f"Lookup VirusTotal {fmt_name} for {entity}."
    return _FUNC_DOC.format(description=description)


def _get_relationships(vt_client, entity_id, vt_type, relationship):
    result_df = vt_client.lookup_ioc_relationships(
        observable=entity_id, vt_type=vt_type, relationship=relationship
    )
    return result_df.reset_index()


def _get_vt_api_scope() -> VTAPIScope:
    """Retrieve the VT enterprise key from settings."""
    prov_settings = get_provider_settings("TIProviders")
    vt_settings = prov_settings.get("VirusTotal")
    if vt_settings:
        return (
            VTAPIScope.PRIVATE
            if vt_settings.args.get(_USE_PRIVATE_API_KEY, False)
            else VTAPIScope.PUBLIC
        )
    return VTAPIScope.PUBLIC



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Bookmark Features."""
from typing import Dict, List, Optional, Union
from uuid import UUID, uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import _build_sent_data, get_http_timeout

__version__ = VERSION
__author__ = "Pete Bryan"


class SentinelBookmarksMixin:
    """Mixin class with Sentinel Bookmark integrations."""

    def list_bookmarks(self) -> pd.DataFrame:
        """
        Return a list of Bookmarks from a Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A set of bookmarks.

        """
        return self._list_items(item_type="bookmarks")  # type: ignore

    def create_bookmark(
        self,
        name: str,
        query: str,
        results: str = None,
        notes: str = None,
        labels: List[str] = None,
    ) -> Optional[str]:
        """
        Create a bookmark in the Sentinel Workspace.

        Parameters
        ----------
        name : str
            The name of the bookmark to use
        query : str
            The KQL query for the bookmark
        results : str, optional
            The results of the query to include with the bookmark, by default None
        notes : str, optional
            Any notes you want associated with the bookmark, by default None
        labels : List[str], optional
            Any labels you want associated with the bookmark, by default None

        Returns
        -------
        Optional[str]
            The name/ID of the bookmark.

        Raises
        ------
        CloudError
            If API returns an error.

        """
        self.check_connected()  # type: ignore
        # Generate or use resource ID
        bkmark_id = str(uuid4())
        bookmark_url = self.sent_urls["bookmarks"] + f"/{bkmark_id}"  # type: ignore
        data_items: Dict[str, Union[str, List]] = {
            "displayName": name,
            "query": query,
        }
        if results:
            data_items["queryResult"] = results
        if notes:
            data_items["notes"] = notes
        if labels:
            data_items["labels"] = labels
        data = _build_sent_data(data_items, props=True)
        params = {"api-version": "2020-01-01"}
        response = httpx.put(
            bookmark_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code == 200:
            print("Bookmark created.")
            return response.json().get("name")
        raise CloudError(response=response)

    def delete_bookmark(
        self,
        bookmark: str,
    ):
        """
        Delete the selected bookmark.

        Parameters
        ----------
        bookmark: str, optional
            The name or GIUD of the bookmark to delete.

        Raises
        ------
        CloudError
            If the API returns an error.

        """
        self.check_connected()  # type: ignore
        bookmark_id = self._get_bookmark_id(bookmark)
        bookmark_url = self.sent_urls["bookmarks"] + f"/{bookmark_id}"  # type: ignore
        params = {"api-version": "2020-01-01"}
        response = httpx.delete(
            bookmark_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            timeout=get_http_timeout(),
        )
        if response.status_code == 200:
            print("Bookmark deleted.")
        else:
            raise CloudError(response=response)

    def _get_bookmark_id(self, bookmark: str) -> str:
        """
        Get the ID of a bookmark.

        Parameters
        ----------
        bookmark : str
            GUID or name of a bookmark

        Returns
        -------
        str
            Bookmark GUID

        Raises
        ------
        MsticpyUserError
            If Bookmark not found or multiple matching bookmarks found.

        """
        try:
            UUID(bookmark)
            return bookmark
        except ValueError as bkmark_name:
            bookmarks = self.list_bookmarks()
            filtered_bookmarks = bookmarks[
                bookmarks["properties.displayName"].str.contains(bookmark)
            ]
            if len(filtered_bookmarks) > 1:
                display(filtered_bookmarks[["name", "properties.displayName"]])
                raise MsticpyUserError(
                    "More than one incident found, please specify by GUID"
                ) from bkmark_name
            if (
                not isinstance(filtered_bookmarks, pd.DataFrame)
                or filtered_bookmarks.empty
            ):
                raise MsticpyUserError(
                    f"Incident {bookmark} not found"
                ) from bkmark_name
            return filtered_bookmarks["name"].iloc[0]

    get_bookmarks = list_bookmarks



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Microsoft Sentinel APIs to interact with Microsoft Sentinel Workspaces."""

import contextlib
import logging
from typing import Any, Dict, List, Optional

import pandas as pd

from ..._version import VERSION
from ...common.exceptions import MsticpyUserConfigError
from ...common.wsconfig import WorkspaceConfig
from .azure_data import AzureData, get_token
from .sentinel_analytics import SentinelAnalyticsMixin, SentinelHuntingMixin
from .sentinel_bookmarks import SentinelBookmarksMixin
from .sentinel_dynamic_summary import SentinelDynamicSummaryMixin, SentinelQueryProvider
from .sentinel_incidents import SentinelIncidentsMixin
from .sentinel_search import SentinelSearchlistsMixin
from .sentinel_ti import SentinelTIMixin
from .sentinel_utils import (
    _PATH_MAPPING,
    SentinelUtilsMixin,
    parse_resource_id,
    validate_res_id,
)
from .sentinel_watchlists import SentinelWatchlistsMixin
from .sentinel_workspaces import SentinelWorkspacesMixin

__version__ = VERSION
__author__ = "Pete Bryan"

logger = logging.getLogger(__name__)


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class MicrosoftSentinel(
    SentinelAnalyticsMixin,
    SentinelHuntingMixin,
    SentinelBookmarksMixin,
    SentinelDynamicSummaryMixin,
    SentinelIncidentsMixin,
    SentinelUtilsMixin,
    SentinelWatchlistsMixin,
    SentinelSearchlistsMixin,
    SentinelWorkspacesMixin,
    SentinelTIMixin,
    AzureData,
):
    """Class for returning key Microsoft Sentinel elements."""

    def __init__(
        self,
        res_id: Optional[str] = None,
        connect: Optional[bool] = False,
        cloud: Optional[str] = None,
        sub_id: Optional[str] = None,
        res_grp: Optional[str] = None,
        ws_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize connector for Azure APIs.

        Parameters
        ----------
        res_id : str, optional
            Set the Sentinel workspace resource ID you want to use, if not specified
            defaults will be looked for or details can be passed separately with functions.
        connect : bool, optional
            Set true if you want to connect to API on initialization, by default False
        cloud : str, optional
            Specify cloud to use, overriding any configuration value.
            Default is to use configuration setting or public cloud if no
            configuration setting is available.
        sub_id : str, optional
            If not specifying a resource ID the Subscription ID of the Sentinel Workspace
            by default None
        res_grp : str, optional
            If not specifying a resource ID the Resource Group name of the
            Sentinel Workspace, by default None
        ws_name : str, optional
            If not specifying a resource ID, the Workspace name of the
            Sentinel Workspace, by default None
        workspace : str, optional
            Alias of ws_name

        """
        super().__init__(connect=False, cloud=cloud)
        self.base_url = self.az_cloud_config.resource_manager
        self.default_subscription: Optional[str] = None
        self._resource_id = res_id
        self._default_resource_group: Optional[str] = None
        self.sent_urls: Dict[str, str] = {}
        self.sent_data_query: Optional[SentinelQueryProvider] = None  # type: ignore
        self.url: Optional[str] = None
        self._token: Optional[str] = None

        workspace = kwargs.get("workspace", ws_name)
        self._default_workspace: Optional[str] = workspace
        self.workspace_config = WorkspaceConfig(workspace)

        logger.info("Initializing Microsoft Sentinel connector")
        logger.info(
            "Params: Cloud=%s; ResourceId=%s; Workspace=%s",
            self.cloud,
            self._resource_id,
            workspace,
        )

        if self._resource_id:
            # If a resource ID is supplied, use that
            logger.info("Initializing from resource ID")
            self.url = self._build_sent_paths(
                self._resource_id, self.base_url
            )  # type: ignore
            res_id_parts = parse_resource_id(self._resource_id)
            self.default_subscription = res_id_parts["subscription_id"]
            self._default_resource_group = res_id_parts["resource_group"]
            self._default_workspace = workspace or res_id_parts["workspace_name"]
            if self._default_workspace in WorkspaceConfig.list_workspaces():
                self.workspace_config = WorkspaceConfig(
                    workspace=self._default_workspace
                )
                logger.info("Workspace settings found for %s", self._default_workspace)
        else:
            # Otherwise - use details from specified workspace or default from settings
            logger.info("Initializing from workspace settings")
            self.default_subscription = self.workspace_config.get(
                "subscription_id", sub_id
            )
            self._default_resource_group = self.workspace_config.get(
                "resource_group", res_grp
            )
            workspace_name = self.workspace_config.get("workspace_name", workspace)
            self._resource_id = self._build_sent_res_id(
                sub_id=self.default_subscription,
                res_grp=self._default_resource_group,
                ws_name=workspace_name,
            )
            logger.info("Resource ID set to %s", self._resource_id)
            self._default_workspace = workspace_name
            self.url = self._build_sent_paths(
                self._resource_id, self.base_url  # type: ignore
            )

        if connect:
            self.connect(**kwargs)

    def connect(
        self,
        auth_methods: Optional[List] = None,
        tenant_id: Optional[str] = None,
        silent: bool = False,
        **kwargs,
    ):
        """
        Authenticate with the SDK & API.

        Parameters
        ----------
        auth_methods : List, optional
            list of preferred authentication methods to use, by default None
        tenant_id : str, optional
            Specify cloud tenant to use
        silent : bool, optional
            Set true to prevent output during auth process, by default False
        cloud : str, optional
            What Azure cloud to connect to.
            By default it will attempt to use the cloud setting from config file.
            If this is not set it will default to Azure Public Cloud
        credential: AzureCredential, optional
            Credentials to use for authentication. This will use the credential
            directly and bypass the MSTICPy Azure credential selection process.

        See Also
        --------
        msticpy.auth.azure_auth.az_connect : function to authenticate to Azure SDK

        """
        if workspace := kwargs.get("workspace"):
            # override any previous default setting
            self.workspace_config = WorkspaceConfig(workspace)
            logger.info("Using workspace settings found for %s", workspace)
        if not self.workspace_config:
            self.workspace_config = WorkspaceConfig()
            logger.info(
                "Using default workspace settings for %s",
                self.workspace_config.get(WorkspaceConfig.CONF_WS_NAME_KEY),
            )
        if kwargs.get("cloud", self.cloud) != self.cloud:
            raise MsticpyUserConfigError(
                "Cannot switch to different cloud",
                f"Current cloud '{self.cloud}'",
                f"Create a new instance of `{self.__class__.__name__}`",
                "and specify the new cloud name using the `cloud` parameter.",
                title="Cannot switch cloud at connect time",
            )
        tenant_id = tenant_id or self.workspace_config[WorkspaceConfig.CONF_TENANT_ID]
        logger.info("Using tenant id %s", tenant_id)
        self._token = kwargs.pop("token", None)
        super().connect(
            auth_methods=auth_methods, tenant_id=tenant_id, silent=silent, **kwargs
        )
        if not self._token:
            logger.info("Getting token for %s", tenant_id)
            self._token = get_token(
                self.credentials, tenant_id=tenant_id, cloud=self.cloud  # type: ignore
            )

        with contextlib.suppress(KeyError):
            logger.info(
                "Setting default subscription to %s from workspace settings",
                self.default_subscription,
            )
            self.default_subscription = self.workspace_config[
                WorkspaceConfig.CONF_SUB_ID_KEY
            ]
            self.set_default_workspace(
                self.default_subscription,
                self.workspace_config[WorkspaceConfig.CONF_WS_NAME_KEY],
            )
        self._create_api_paths_for_workspace(
            az_resource_id=None,
            subscription_id=self.workspace_config.get(WorkspaceConfig.CONF_SUB_ID_KEY),
            resource_group=self.workspace_config.get(
                WorkspaceConfig.CONF_RES_GROUP_KEY
            ),
            workspace_name=self.workspace_config.get(WorkspaceConfig.CONF_WS_NAME_KEY),
        )

    def _create_api_paths_for_workspace(
        self,
        az_resource_id: Optional[str] = None,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        workspace_name: Optional[str] = None,
    ):
        """Save configuration and build API URLs for workspace."""
        if workspace_name:
            self.workspace_config = WorkspaceConfig(workspace=workspace_name)
        az_resource_id = (
            az_resource_id
            or self._resource_id
            or self._build_sent_res_id(
                subscription_id, resource_group, workspace_name  # type: ignore
            )
        )
        az_resource_id = validate_res_id(az_resource_id)
        self.url = self._build_sent_paths(az_resource_id, self.base_url)  # type: ignore

        self.sent_urls = {
            name: f"{self.url}{mapping}" for name, mapping in _PATH_MAPPING.items()
        }
        logger.info("API URLs set to %s", self.sent_urls)

    def set_default_subscription(self, subscription_id: str):
        """Set the default subscription to use to `subscription_id`."""
        subs_df = self.get_subscriptions()
        logger.info("Setting default subscription to %s", subscription_id)
        if subscription_id in subs_df["Subscription ID"].values:
            self.default_subscription = subscription_id
        else:
            print(f"Subscription ID {subscription_id} not found.")
            print(
                f"Subscriptions found: {', '.join(subs_df['Subscription ID'].values)}"
            )

    def set_default_workspace(
        self, sub_id: Optional[str], workspace: Optional[str] = None
    ):
        """
        Set the default workspace.

        Parameters
        ----------
        sub_id : Optional[str], optional
            Subscription ID containing the workspace. If not specified,
            the subscription will be taken from the `default_subscription`
            or from configuration.
        workspace : Optional[str], optional
            Name of the workspace, by default None.
            If not specified and there is only one workspace in the
            subscription, this will be set as the default.

        Raises
        ------
        ValueError
            If no current or default subscription has been set.

        """
        sub_id = sub_id or self.default_subscription
        if not sub_id:
            raise MsticpyUserConfigError(
                "No current or default subscription ID set.",
                "Please configure the subscription ID for your workspace in your"
                "msticpyconfig.yaml",
            )
        self._default_workspace = workspace
        ws_res_id: Optional[str] = None
        # if workspace not supplied trying looking up in subscription
        if not workspace:
            logger.info("Trying to set default workspace from subscription %s", sub_id)
            workspaces = self.get_sentinel_workspaces(sub_id=sub_id)
            if len(workspaces) == 1:
                # if only one, use that one
                name, ws_res_id = next(iter(workspaces.items()))
                self._default_workspace = name

        # if workspace is one that we have configuration for, get the details from there.
        if self._default_workspace in WorkspaceConfig.list_workspaces():
            logger.info("Workspace %s found in settings", self._default_workspace)
            self.workspace_config = WorkspaceConfig(workspace=self._default_workspace)
        elif ws_res_id:
            # otherwise construct partial settings
            res_id_parts = parse_resource_id(ws_res_id)
            self.workspace_config = WorkspaceConfig.from_settings(
                {
                    "WorkspaceName": (
                        self._default_workspace or res_id_parts["workspace_name"]
                    ),
                    "SubscriptionId": res_id_parts["subscription_id"],
                    "ResourceGroup": res_id_parts["resource_group"],
                }
            )
            logger.info(
                "Workspace not found in settings, using partial workspace config %s",
                self.workspace_config,
            )

    @property
    def default_workspace_settings(self) -> Optional[Dict[str, Any]]:
        """Return current default workspace settings."""
        return self.workspace_config.mp_settings

    @property
    def default_workspace_name(self):
        """Return the default workspace name."""
        return self._default_workspace

    def list_data_connectors(self) -> pd.DataFrame:
        """
        List deployed data connectors.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the deployed data connectors

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        return self._list_items(item_type="data_connectors")


# Alias for old class name
AzureSentinel = MicrosoftSentinel
